/*
    This file is part of Frozen Bubble for S60.

    Frozen Bubble for S60 is free software; you can redistribute
    it and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    Frozen Bubble for S60 is distributed in the hope that it
    will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Frozen Bubble for S60; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <e32math.h>

#ifdef UIQ3
	#include<devicekeys.h>
	#include <eikappui.h>
#else
	#include <AknUtils.h>
#endif

#include <eikapp.h>
#include <barsread.h>
#include "CFrozenBubbleMain.h"
#include "CExtendedBitmap.h"
#include "CExtendedFont.h"
#include "ShootData.h"
#include <balls.mbg>
#include <root.mbg>
#include <menu.mbg>
#include <shoot.mbg>
#include <pinguins.mbg>
#include "CSndMixer.h"
#include "CWavLoader.h"

// bubble types
const TInt BUBBLE_NULL    = -1;
const TInt BUBBLE_BLACK   = 0;
const TInt BUBBLE_WHITE   = 1;
const TInt BUBBLE_BLUE    = 2;
const TInt BUBBLE_GREEN   = 3;
const TInt BUBBLE_YELLOW  = 4;
const TInt BUBBLE_MAGENTA = 5;
const TInt BUBBLE_RED     = 6;
const TInt BUBBLE_ORANGE  = 7;

// Ini file stream ids
const TUid uidSettings  = { 1 };
const TUid uidHiscores  = { 2 };
const TUid uidSavedGame = { 3 };

// Settings record
struct TSettingsRecord{
  TStdScanCode keys[4];
  TBool bSound;
};

struct TSavedGameRecord{
  TInt pdataScore, pdataNewroot, pdataNewroot_prelight, pdataOddswap, pdataMalus, pdataHurry, pdataNewrootlevel, pdataNewroot_prelight_step;
  TInt pdataPing_rightImg, pdataPing_rightState, pdataPing_rightMovelatency, pdataState, pdataHurry_oddness, 
       pdataSticking_step, pdataSleeping, pdataPing_rightActionlatency, pdataSticking_step_slowdown;
  TInt pinguinWin_roll_back_index, pinguinLose_roll_back_index;
  TReal angle;
  TUint time_1pgameextra;
  TInt LevelsCurrent;
  TBubble next_bubble, tobe_launched;
  TBubble sticked_bubbles[14*7];
};

// global states
const TInt GS_MENU       = 1;
const TInt GS_GAME       = 2;
const TInt GS_NAME       = 3;
const TInt GS_HIGHSCORES = 4;
const TInt GS_KEYS       = 5;
const TInt GS_PAUSE      = 6;
const TInt GS_LOADGAME   = 7;
const TInt GS_SAVEGAME   = 8;

const TInt TIME_APPEARS_NEW_ROOT = 8;
const TInt TIME_HURRY_WARN = 400;
const TInt TIME_HURRY_MAX = 525;
const TInt TIMEOUT_PINGUIN_SLEEP = 200;

// speed constants
const TInt TARGET_ANIM_SPEED = 31;  // number of milliseconds that should last between two animation frames
const TReal LAUNCHER_SPEED = 0.03;  // speed of rotation of launchers
const TReal BUBBLE_SPEED = 6;       // speed of movement of launched bubbles
const TReal FREE_FALL_CONSTANT = 0.5;

// ping_right states
const TInt STATE_NORMAL = 1;
const TInt STATE_GAME   = 2;
const TInt STATE_LOST   = 3;
const TInt STATE_WON    = 4;
const TInt STATE_SLEEP  = 5;
const TInt STATE_LEFT   = 6;
const TInt STATE_RIGHT  = 7;
const TInt STATE_ACTION = 8;


// simple macros
inline TBool odd(TInt v){
  return v % 2 == 1;
}
inline TBool even(TInt v){
  return v % 2 == 0;
}
inline TReal sqr(TReal v){
  return v*v;
}
inline TReal abs(TReal v){
  return v > 0 ? v : -v;
}
inline TReal sqrt(TReal v){
  TReal res;
  Math::Sqrt(res, v);
  return res;
}
inline TBool log_xor(TInt a, TInt b){
  return (a!=0) != (b!=0);
}
inline TReal cos(TReal val){
  TReal res;
  Math::Cos(res,val);
  return res;
}
inline TReal sin(TReal val){
  TReal res;
  Math::Sin(res,val);
  return res;
}
inline void AppendBubbleArray(RBubbleArray &source, RBubbleArray &dest){
  for(TInt idx=0; idx<source.Count(); idx++)
    dest.Append( source[idx] );
}
inline void CopyBubbleArray(RBubbleArray &source, RBubbleArray &dest){
  dest.Reset();
  AppendBubbleArray(source, dest);
}
inline TInt strlen(TUint8 *str){
  TInt c=0;
  while(str[c] != '\0')
    c++;
  return c;
}

CFrozenBubbleMain::~CFrozenBubbleMain(){
//  smxMain->Resume();
  delete prdGameLoop;
  // Cleanup bitmaps
  delete imgOrigBackground;
  delete imgBack_1p;
  delete imgBackstartfull;
  // delete ball image mask explicitly so not to try to delete it more than once
  TInt idx;
  for(idx=0; idx<8; idx++){
    delete imgBall[idx];
  }
  delete imgBubblesAnimLose;
  for(idx=0; idx<8; idx++)
    delete imgBubblesAnim[idx];
  for(idx=0; idx<(TInt)(sizeof(imgCanon)/sizeof(*imgCanon)); idx++)
    delete imgCanon[idx];
  delete canondata;
  delete imgBack_paused;
  delete imgCompressorExt;
  delete imgCompressorMain;
  delete imgLose;
  delete imgWin;
  for(idx=0; idx<(TInt)(sizeof(imgPause)/sizeof(*imgPause)); idx++)
    delete imgPause[idx];
  delete imgTxt_1pgame_off;
  delete imgTxt_1pgame_over;
  delete imgTxt_keys_off;
  delete imgTxt_keys_over;
  delete imgTxt_sound_off;
  delete imgTxt_sound_over;
  delete imgTxt_sound_act_off;
  delete imgTxt_sound_act_over;
  delete imgTxt_highscores_off;
  delete imgTxt_highscores_over;
  delete imgVoidPanel;
//  delete imgQueryYesNo;
  delete imgBack_hiscores;
  delete imgBannerArtwork;
  delete imgBannerCpucontrol;
  delete imgBannerPorting;
  delete imgBannerSoundtrack;
  delete imgHurry;
  for(idx=0; idx<(TInt)(sizeof(imgPinguinNormal)/sizeof(*imgPinguinNormal)); idx++)
    delete imgPinguinNormal[idx];
  for(idx=0; idx<(TInt)(sizeof(imgPinguinLose)/sizeof(*imgPinguinLose)); idx++)
    delete imgPinguinLose[idx];
  for(idx=0; idx<(TInt)(sizeof(imgPinguinWin)/sizeof(*imgPinguinWin)); idx++)
    delete imgPinguinWin[idx];
  delete imgPinguinAction;
  delete imgPinguinMoveLeft;
  delete imgPinguinMoveRight;
  delete imgPinguinSleep;
  // Font
  delete fntNormal;
  delete fntDark;
  // Other objects
  sticked_bubbles.Close();
  falling_bubble.Close();
  exploding_bubble.Close();
  chains_falling_chained.Close();
  chains_chained_bubbles.Close();
  will_destroy.Close();
  tmp_bubbles.Close();
  
  highscores.Close();
  // sound classes
  delete smxMain;
  delete smpStick.iData;
  delete smpLaunch.iData;
  delete smpDestroy_group.iData;
  delete smpApplause.iData;
  delete smpCancel.iData;
  delete smpHurry.iData;
  delete smpLose.iData;
  delete smpMenu_change.iData;
  delete smpMenu_selected.iData;
  delete smpNewroot_solo.iData;
  delete smpNoh.iData;
  delete smpRebound.iData;
  delete smpSnore.iData;
  delete smpTypewriter.iData;
}

void CFrozenBubbleMain::main(void){
  init_game();
}

TKeyResponse CFrozenBubbleMain::DoKeyEvent(const TKeyEvent& aKeyEvent, TEventCode aType){
  if( aType == EEventKeyDown ){
    sccPressedButton = (TStdScanCode)aKeyEvent.iScanCode;
    return EKeyWasConsumed;
  }
  if( aType == EEventKeyUp ){
    sccPressedButton = EStdKeyNull;
    return EKeyWasConsumed;
  }
  if( aType == EEventKey ){
    sccLastKeyEvent = (TStdScanCode)aKeyEvent.iScanCode;
    return EKeyWasConsumed;
  }
  return EKeyWasNotConsumed;
}

void CFrozenBubbleMain::ExternalEvent(TInt event){
  switch(event){
    case EVT_SELECT:
      switch(global_state){
        case GS_GAME:
          CalcTimer();
          dopause();
          break;
        case GS_MENU:
#ifdef UIQ3
		  sccLastKeyEvent = EStdDeviceKeyAction;
#else
          sccLastKeyEvent = EStdKeyDevice3;
#endif
          break;
        case GS_NAME:
          break;
        case GS_HIGHSCORES:
          break;
        case GS_KEYS:
          break;
        case GS_PAUSE:
          break;
        case GS_LOADGAME:
          InitTimer();
          new_game(ETrue);
          break;
        case GS_SAVEGAME:
          {
            CPrintableExtendedBitmap* imgTemp;
            imgTemp = new(ELeave)CPrintableExtendedBitmap;
            CleanupStack::PushL(imgTemp);
            CFbsBitmap* iBitmap = new ( ELeave ) CFbsBitmap();
            CleanupStack::PushL( iBitmap );
            User::LeaveIfError( iBitmap->Create( EikonEnv->ScreenDevice()->SizeInPixels(), EikonEnv->DefaultDisplayMode() ) );
            imgTemp->SetBitmap( iBitmap );
            Actions->BitBltBack( *imgTemp );
            CleanupStack::PopAndDestroy( iBitmap );
//            imgTemp->CopyFrom( *imgBack_1p->GetBitmap() );
            AddBrightness(*imgTemp);
            Actions->BitBlt(BLTTYPE_NORMAL, *imgTemp, ptZero);
            Actions->FlipBackBuffer();
            CleanupStack::PopAndDestroy(imgTemp);
            save_game();
//            Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_paused, ptTopLeft);
//            Actions->FlipBackBuffer();
            // reset LevelsCurrent to avoid placing a new high score
            LevelsCurrent = 0;
            menu();
          }
          break;
      }
      break;
    case EVT_ESCAPE:
      if( global_state == GS_PAUSE ){
        RebuildOriginalBackground();
        ResetTimeStamp();
		smxMain->PauseMusic(EFalse);
        setGameLoop(maingame_);
      } else if( global_state == GS_LOADGAME ){
        InitTimer();
        new_game(EFalse);
      } else if( global_state == GS_SAVEGAME ){
        // erase saved game
        if( bLoadedGame )
          Actions->RemoveIniData(uidSavedGame);
        Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_1p, ptZero); 
        menu();
      } else if( global_state == GS_GAME ){
        CalcTimer();
        querysavegame();
      } else if( global_state != GS_MENU ){
        if( global_state == GS_NAME )
          play_sound(smpCancel);        
        menu();
      } else
        Actions->DoExit();
      break;
    case EVT_GOBACKGR:
      if(bSoundEnabled && (global_state!=0))
        smxMain->Pause();
      setGameLoop(NULL);
      CalcTimer();
      break;
    case EVT_GOFOREGR:
      if(bSoundEnabled && (global_state!=0))
        smxMain->Resume();
      switch(global_state){
        case GS_GAME:
          dopause();
          break;
        case GS_MENU:
          setGameLoop(menu_loop_);
          break;
        case GS_NAME:
          setGameLoop(ask_nameloop_);
          break;
        case GS_HIGHSCORES:
          setGameLoop(highscoreloop_);
          break;
        case GS_KEYS:
          setGameLoop(ask_keysloop_);
          break;
        case GS_PAUSE:
          setGameLoop(dopause_loop_);
          break;
        case GS_LOADGAME:
        case GS_SAVEGAME:
          break;
      }
      break;
  }
}

void CFrozenBubbleMain::ForceQuit(void){
  switch(global_state){
    case GS_GAME:
    case GS_PAUSE:
    case GS_SAVEGAME:
      save_game();
      break;
    default:
      break;
  }
}

void CFrozenBubbleMain::init_game(void){
  init_game_rest();
  init_game_graphics();
  init_game_sound();
  menu();
}

void CFrozenBubbleMain::init_game_graphics(void){
  _LIT (KMenuImagesFilename,    "menu.mbm");
  _LIT (KRootImagesFilename,    "root.mbm");
  _LIT (KBallsImagesFilename,   "balls.mbm");
  _LIT (KCanonImagesFilename,   "shoot.mbm");
  _LIT (KPinguinImagesFilename, "pinguins.mbm");

  TDisplayMode dmMain = EikonEnv->DefaultDisplayMode();

  CFbsBitmap* iBitmap = new ( ELeave ) CFbsBitmap();
  CleanupStack::PushL( iBitmap );
  CExtendedBitmap *imgSplash;
  InitializeBitmap(&imgSplash,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuSplash), dmMain), ETrue,
                   NULL, EFalse);
  
  Actions->BitBlt(BLTTYPE_NORMAL, *imgSplash, ptZero);
  Actions->FlipBackBuffer();
  imgOrigBackground = new (ELeave) CPrintableExtendedBitmap();
  imgOrigBackground->SetBitmap( imgSplash->GetBitmap() );
  User::LeaveIfError( iBitmap->Create( EikonEnv->ScreenDevice()->SizeInPixels(), dmMain ) );
  AddBrightness(*imgOrigBackground);
  Actions->BitBlt(BLTTYPE_NORMAL, *imgOrigBackground, ptZero);  
  imgOrigBackground->SetBitmap( iBitmap );
  delete imgSplash;
  imgSplash = NULL;  
  CleanupStack::PopAndDestroy( iBitmap );
  iBitmap = NULL;

  InitializeBitmap(&imgBack_1p,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootBack_one_player), dmMain ), ETrue,
                   NULL, EFalse);
  InitializeBitmap(&imgBackstartfull,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBack_start), dmMain), ETrue,
                   NULL, EFalse);

  // Load ball pictures
  iBitmap = EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_mask);
  InitializeBitmap(&imgBall[0],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_1), dmMain ), ETrue,
                   iBitmap, EFalse);
  InitializeBitmap(&imgBall[1],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_2), dmMain ), ETrue,
                   iBitmap, EFalse);
  InitializeBitmap(&imgBall[2],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_3), dmMain ), ETrue,
                   iBitmap, EFalse);
  InitializeBitmap(&imgBall[3],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_4), dmMain ), ETrue,
                   iBitmap, EFalse);
  InitializeBitmap(&imgBall[4],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_5), dmMain ), ETrue,
                   iBitmap, EFalse);
  InitializeBitmap(&imgBall[5],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_6), dmMain ), ETrue,
                   iBitmap, EFalse);
  InitializeBitmap(&imgBall[6],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_7), dmMain ), ETrue,
                   iBitmap, EFalse);
  InitializeBitmap(&imgBall[7],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_8), dmMain ), ETrue,
                   iBitmap, EFalse);
  delete iBitmap;


  InitializeBitmap(&imgBubblesAnimLose,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_lose), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsBubble_lose_msk), ETrue);

  InitializeBitmap(&imgBubblesAnim[0],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_0), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_0_msk), ETrue);
  InitializeBitmap(&imgBubblesAnim[1],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_1_msk), ETrue);
  InitializeBitmap(&imgBubblesAnim[2],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_2), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_2_msk), ETrue);
  InitializeBitmap(&imgBubblesAnim[3],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_3), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_3_msk), ETrue);
  InitializeBitmap(&imgBubblesAnim[4],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_4), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_4_msk), ETrue);
  InitializeBitmap(&imgBubblesAnim[5],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_5), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_5_msk), ETrue);
  InitializeBitmap(&imgBubblesAnim[6],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_6), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_6_msk), ETrue);
  InitializeBitmap(&imgBubblesAnim[7],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_7), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KBallsImagesFilename, EMbmBallsStick_effect_7_msk), ETrue);

  InitializeBitmap(&imgBack_paused,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootBack_paused), dmMain ), ETrue,
                   NULL, EFalse);

  InitializeBitmap(&imgCompressorExt,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootCompressor_ext), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootCompressor_ext_msk), ETrue);

  InitializeBitmap(&imgCompressorMain,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootCompressor_main), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootCompressor_main_msk), ETrue);

  InitializeBitmap(&imgLose,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootLose_panel), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootLose_panel_msk), ETrue);

  InitializeBitmap(&imgWin,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootWin_panel_1player), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootWin_panel_1player_msk), ETrue);

  InitializeBitmap(&imgPause[0],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootPause_1), dmMain ), ETrue,
                   NULL, EFalse);
  InitializeBitmap(&imgPause[1],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootPause_2), dmMain ), ETrue,
                   NULL, EFalse);
  InitializeBitmap(&imgPause[2],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootPause_3), dmMain ), ETrue,
                   NULL, EFalse);
  InitializeBitmap(&imgPause[3],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootPause_4), dmMain ), ETrue,
                   NULL, EFalse);
  InitializeBitmap(&imgPause[4],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootPause_5), dmMain ), ETrue,
                   NULL, EFalse);

  // "Loading" bitmap
  CExtendedBitmap *imgLoading;
  InitializeBitmap(&imgLoading,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootLoading), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootLoading_msk), ETrue);
  Actions->BitBlt(BLTTYPE_MASKED, *imgLoading, TPoint(70, 148));
  
  delete imgLoading;
  Actions->FlipBackBuffer();

  // Load cannon bitmaps
  TInt idx, idx2;
  for(idx2=EMbmShootBase_m40, idx=0; idx<(TInt)(sizeof(imgCanon)/sizeof(*imgCanon)); idx++){
    CFbsBitmap *iCanon = ConvertBitmapL( EikonEnv->CreateBitmapL(KCanonImagesFilename, idx2++), dmMain );
    CFbsBitmap *iCanonAlpha = EikonEnv->CreateBitmapL(KCanonImagesFilename, idx2++);
    InitializeBitmap(&imgCanon[idx],
                     iCanon, ETrue,
                     iCanonAlpha, ETrue);
  }

  canondata = new(ELeave) TShootData;
  canondata->InitL();
  
  InitializeBitmap(&imgTxt_1pgame_off,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_1pgame_off), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_1pgame_off_msk), ETrue);
  
  InitializeBitmap(&imgTxt_1pgame_over,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_1pgame_over), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_1pgame_over_msk), ETrue);
  
  InitializeBitmap(&imgTxt_keys_off,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_keys_off), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_keys_off_msk), ETrue);

  InitializeBitmap(&imgTxt_keys_over,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_keys_over), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_keys_over_msk), ETrue);

  InitializeBitmap(&imgTxt_sound_off,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_off), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_off_msk), ETrue);

  InitializeBitmap(&imgTxt_sound_over,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_over), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_over_msk), ETrue);

  InitializeBitmap(&imgTxt_sound_act_off,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_act_off), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_act_off_msk), ETrue);

  InitializeBitmap(&imgTxt_sound_act_over,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_act_over), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_sound_act_over_msk), ETrue);

  InitializeBitmap(&imgTxt_highscores_off,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_highscores_off), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_highscores_off_msk), ETrue);

  InitializeBitmap(&imgTxt_highscores_over,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_highscores_over), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuTxt_highscores_over_msk), ETrue);

  InitializeBitmap(&imgVoidPanel,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuVoid_panel), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuVoid_panel_msk), ETrue);

/*  InitializeBitmap(&imgQueryYesNo,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootQueryyesno), dmMain ), ETrue,
                   NULL, EFalse);*/

  InitializeBitmap(&imgBack_hiscores,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootBack_hiscores), dmMain ), ETrue,
                   NULL, EFalse);

  InitializeBitmap(&imgBannerArtwork,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_artwork), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_artwork_msk), ETrue);
  InitializeBitmap(&imgBannerCpucontrol,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_cpucontrol), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_cpucontrol_msk), ETrue);
  InitializeBitmap(&imgBannerPorting,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_porting), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_porting_msk), ETrue);

  InitializeBitmap(&imgBannerSoundtrack,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_soundtrack), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KMenuImagesFilename, EMbmMenuBanner_soundtrack_msk), ETrue);

  InitializeBitmap(&imgHurry,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootHurry_p1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootHurry_p1_msk), ETrue);
  InitializeBitmap(&imgPinguinNormal[0],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1_msk), ETrue);
  InitializeBitmap(&imgPinguinNormal[1],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1_extra_01), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1_extra_01_msk), ETrue);
  InitializeBitmap(&imgPinguinNormal[2],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1_extra_02), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1_extra_02_msk), ETrue);
  InitializeBitmap(&imgPinguinNormal[3],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1_extra_03), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsBase_p1_extra_03_msk), ETrue);
  
  InitializeBitmap(&imgPinguinAction,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsAction_p1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsAction_p1_msk), ETrue);
  
  InitializeBitmap(&imgPinguinMoveLeft,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsMove_left_p1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsMove_left_p1_msk), ETrue);

  InitializeBitmap(&imgPinguinMoveRight,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsMove_right_p1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsMove_right_p1_msk), ETrue);

  InitializeBitmap(&imgPinguinSleep,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsSleep_p1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsSleep_p1_msk), ETrue);

  
  pinguinWin_roll_back_index = 4;

  InitializeBitmap(&imgPinguinLose[0],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_1_msk), ETrue);
  
  InitializeBitmap(&imgPinguinLose[1],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_2), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_2_msk), ETrue);

  InitializeBitmap(&imgPinguinLose[2],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_3), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_3_msk), ETrue);

  InitializeBitmap(&imgPinguinLose[3],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_4), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_4_msk), ETrue);

  InitializeBitmap(&imgPinguinLose[4],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_5), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_5_msk), ETrue);

  InitializeBitmap(&imgPinguinLose[5],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_6), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_6_msk), ETrue);

  InitializeBitmap(&imgPinguinLose[6],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_7), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_7_msk), ETrue);

  InitializeBitmap(&imgPinguinLose[7],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_8), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_8_msk), ETrue);

  InitializeBitmap(&imgPinguinLose[8],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_9), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_loose_9_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[0],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_1), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_1_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[1],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_2), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_2_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[2],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_3), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_3_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[3],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_4), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_4_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[4],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_5), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_5_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[5],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_6), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_6_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[6],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_7), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_7_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[7],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_8), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_8_msk), ETrue);

  InitializeBitmap(&imgPinguinWin[8],
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_6), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KPinguinImagesFilename, EMbmPinguinsP1_win_6_msk), ETrue);

  CExtendedBitmap *imgFont;// = new(ELeave) CExtendedBitmap();
  InitializeBitmap(&imgFont,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootFont), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootFont_msk), ETrue);

  // load fonts  
  fntNormal = new(ELeave) CExtendedFont(imgFont, Actions);
  fntNormal->ConstructL();
  
  InitializeBitmap(&imgFont,
                   ConvertBitmapL( EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootFont_hi), dmMain ), ETrue,
                   EikonEnv->CreateBitmapL(KRootImagesFilename, EMbmRootFont_hi_msk), ETrue);
  fntDark = new(ELeave) CExtendedFont(imgFont, Actions); 
  fntDark->ConstructL();
}

void CFrozenBubbleMain::init_game_sound(void){
  // Sound classes initialization
  CWavLoader *wldMain = CWavLoader::NewLC();
  smxMain = CSndMixer::NewL();
  if(bSoundEnabled)
    smxMain->Resume();
  TFileName filename;
  filename = _L("stick.snd");
  load_sound_sample(*wldMain, smpStick, filename);//_L("stick.snd"));
  filename = _L("launch.snd");
  load_sound_sample(*wldMain, smpLaunch, filename);//_L("launch.snd"));
  filename = _L("destroy_group.snd");
  load_sound_sample(*wldMain, smpDestroy_group, filename);//_L("destroy_group.snd"));
  filename = _L("applause.snd");
  load_sound_sample(*wldMain, smpApplause, filename);//_L("applause.snd"));
  filename = _L("cancel.snd");
  load_sound_sample(*wldMain, smpCancel, filename);//_L("cancel.snd"));
  filename = _L("hurry.snd");
  load_sound_sample(*wldMain, smpHurry, filename);//_L("hurry.snd"));
  filename = _L("lose.snd");
  load_sound_sample(*wldMain, smpLose, filename);//_L("lose.snd"));
  filename = _L("menu_change.snd");
  load_sound_sample(*wldMain, smpMenu_change, filename);//_L("menu_change.snd"));
  filename = _L("menu_selected.snd");
  load_sound_sample(*wldMain, smpMenu_selected, filename);//_L("menu_selected.snd"));
  filename = _L("newroot_solo.snd");
  load_sound_sample(*wldMain, smpNewroot_solo, filename);//_L("newroot_solo.snd"));
  filename = _L("noh.snd");
  load_sound_sample(*wldMain, smpNoh, filename);//_L("noh.snd"));
  filename = _L("rebound.snd");
  load_sound_sample(*wldMain, smpRebound, filename);//_L("rebound.snd"));
  filename = _L("snore.snd");
  load_sound_sample(*wldMain, smpSnore, filename);//_L("snore.snd"));
  filename = _L("typewriter.snd");
  load_sound_sample(*wldMain, smpTypewriter, filename);//_L("typewriter.snd"));
  CleanupStack::PopAndDestroy( wldMain );
}

void CFrozenBubbleMain::load_sound_sample(CWavLoader& loader, TSample& sample, const TFileName& filename){
	TBuf<100> iSoundFile(iPrivatePath);
	iSoundFile.Append(filename);
  sample = loader.LoadL( iSoundFile );
}

void CFrozenBubbleMain::init_game_rest(void){
 EikonEnv->FsSession().PrivatePath(iPrivatePath);
#ifdef __WINS__
  iPrivatePath.Insert(0, _L("C:"));
#else
  TFileName appFullName = EikonEnv->EikAppUi()->Application()->AppFullName();
  TParsePtrC parse(appFullName);
  iPrivatePath.Insert(0, parse.Drive());
#endif

  iScreenRes = EikonEnv->ScreenDevice()->SizeInPixels();
  
  prdGameLoop = CPeriodic::NewL(CActive::EPriorityIdle);

  total_time.HomeTime();

  iRandomSeed = getRandomSeed();
  
//  highscores = new(ELeave) RHighScores;
#ifdef UIQ3
  KEYS_left   = EStdDeviceKeyFourWayLeft;
  KEYS_right  = EStdDeviceKeyFourWayRight;
  KEYS_fire   = EStdDeviceKeyAction;
  KEYS_center = EStdDeviceKeyFourWayUp;

#else
  KEYS_left   = EStdKeyLeftArrow;
  KEYS_right  = EStdKeyRightArrow;
  KEYS_fire   = EStdKeyDevice3;
  KEYS_center = EStdKeyUpArrow;
#endif
  
  pdataScore = 0;
  load_levelset();

  bSoundEnabled = ETrue;
  
  load_settings();
  load_hiscores();
}

void CFrozenBubbleMain::menu(){
	play_music(_L8("introzik.it"));

	if( handle_new_hiscores() )
     return;
  
  iMenuCurrent_pos = 0;
  BANNER_START = 380/*720*/; BANNER_SPACING = 40;//80;
  const TInt TOTAL_BANNERS = sizeof(BANNERS)/sizeof(*BANNERS);
  BANNER_MINX = 58; BANNER_MAXX = 202; BANNER_Y = 298;
  BANNER_RECT.SetRect( TPoint(BANNER_MINX, BANNER_Y), TSize(BANNER_MAXX-BANNER_MINX, 30) );
  TSize szBannerPorting = imgBannerPorting->GetBitmap()->SizeInPixels();
  TSize szBannerArtwork = imgBannerArtwork->GetBitmap()->SizeInPixels();
  TSize szBannerSoundtrack = imgBannerSoundtrack->GetBitmap()->SizeInPixels();
  BANNERS[0] = BANNER_START;
  BANNERS[1] = BANNER_START + szBannerPorting.iWidth + BANNER_SPACING;
  BANNERS[2] = BANNER_START + szBannerPorting.iWidth + BANNER_SPACING + szBannerArtwork.iWidth + BANNER_SPACING ;
  BANNERS[3] = BANNER_START + szBannerPorting.iWidth + BANNER_SPACING + szBannerArtwork.iWidth + BANNER_SPACING + szBannerSoundtrack.iWidth + BANNER_SPACING;

  BANNERS_MAX = BANNERS[TOTAL_BANNERS-1] - (iScreenRes.iWidth - (BANNER_MAXX-BANNER_MINX)) + BANNER_SPACING;
  iBannerPos = 280;//670;

//  sccLastKeyEvent = EStdKeyNull;

  setGameLoop(menu_loop_);
}

TBool CFrozenBubbleMain::handle_new_hiscores() {
  if( !LevelsCurrent ) 
    return EFalse;

  candidateHighScore.level = LevelsCurrent;
  candidateHighScore.seconds = time_1pgameextra / 1000;
  LevelsCurrent = 0;

  if(highscores.Count()==10){
    THighScoreRecord &worst = highscores[highscores.Count()-1];
    if( HighScoreSort(candidateHighScore, worst) >= 0 )
      return EFalse;
  }
  play_sound(smpApplause);
  
  ask_name();
  return ETrue;
}

void CFrozenBubbleMain::ask_name(void) {
//  sccLastKeyEvent = EStdKeyNull;
  TSize szVoidPanel = imgVoidPanel->GetBitmap()->SizeInPixels();
  TInt xpos_panel = (iGameRes.iWidth-szVoidPanel.iWidth)/2;
  TInt ypos_panel = (iGameRes.iHeight-szVoidPanel.iHeight)/2;
  Actions->BitBlt(BLTTYPE_MASKED, *imgVoidPanel, TPoint(xpos_panel, ypos_panel));

  hs_ypos = ypos_panel + 5;
  TUint8 *cCongratulations = (TUint8*)"CONGRATULATIONS!";
  TUint8 *cNewHighScore1   = (TUint8*)"YOU HAVE A";
  TUint8 *cNewHighScore2   = (TUint8*)"HIGHSCORE!";
  fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cCongratulations))/2, hs_ypos, cCongratulations);
  hs_ypos += 22;
  fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cNewHighScore1))/2, hs_ypos, cNewHighScore1);
  hs_ypos += 22;
  fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cNewHighScore2))/2, hs_ypos, cNewHighScore2);
  hs_ypos += 22;
  hs_ypos += 5;  

  TUint8 *cYourName = (TUint8*)"YOUR NAME?";
  
  TInt xpos = (iGameRes.iWidth-szVoidPanel.iWidth)/2 + 10;
  fntNormal->Print(xpos, hs_ypos, cYourName);
  hs_ypos += 22;
  Actions->BitBltBack(*imgOrigBackground);
  Actions->FlipBackBuffer();
  for(TUint idx=0; idx<sizeof(candidateHighScore.name)/sizeof(*candidateHighScore.name); idx++)
    candidateHighScore.name[idx] = '\0';
  editEnded = EFalse;
  
  setGameLoop(ask_nameloop_);
}

TInt CFrozenBubbleMain::ask_nameloop_(TAny* aPtr){
  CFrozenBubbleMain *iMain = (CFrozenBubbleMain*)aPtr;
  iMain->ask_nameloop();
  return 0;
}

void CFrozenBubbleMain::ask_nameloop(void){
  ticks_last_key++;
  const TSize szVoidPanel = imgVoidPanel->GetBitmap()->SizeInPixels();

  TStdScanCode k = sccLastKeyEvent;
  sccLastKeyEvent = EStdKeyNull;
  if(editEnded){
    ticks_last_key++;
    if(ticks_last_key > 60)
      display_highscores();
    return;
  }
  if(k != EStdKeyNull){
    /*
    $k == SDLK_ESCAPE and $ok = 0, last entries;
    */
    play_sound(smpTypewriter);
    TInt hscorelen = strlen((TUint8*)candidateHighScore.name);
    if( (k >= '0') && (k <= '9') )
      if( (ticks_last_key > 35) || (k != sccPreviousKeyEvent) ){
        if( hscorelen < HIGHSCORELIMIT ){
	        switch( (TUint8)k ){
	          case '0' :
	            candidateHighScore.name[hscorelen] = ' ';
	            break;
	          case '1' :
	            candidateHighScore.name[hscorelen] = '1';
	            break;
	          case '2' :
	            candidateHighScore.name[hscorelen] = 'A';
	            break;
	          case '3' :
	            candidateHighScore.name[hscorelen] = 'D';
	            break;
	          case '4' :
	            candidateHighScore.name[hscorelen] = 'G';
	            break;
	          case '5' :
	            candidateHighScore.name[hscorelen] = 'J';
	            break;
	          case '6' :
	            candidateHighScore.name[hscorelen] = 'M';
	            break;
	          case '7' :
	            candidateHighScore.name[hscorelen] = 'P';
	            break;
	          case '8' :
	            candidateHighScore.name[hscorelen] = 'T';
	            break;
	          case '9' :
	            candidateHighScore.name[hscorelen] = 'W';
	            break;
	        }
	        ticks_last_key = 0;	        
        }
      } else if( hscorelen>0 ){
        switch( (TUint8)k ){
          case '0' :
            if( candidateHighScore.name[hscorelen-1] == ' ' )
              candidateHighScore.name[hscorelen-1] = '0';
            else
              candidateHighScore.name[hscorelen-1] = ' ';
            break;
          case '2' :
            if( candidateHighScore.name[hscorelen-1] == '2' )
              candidateHighScore.name[hscorelen-1] = 'A';          
            else if( candidateHighScore.name[hscorelen-1] == 'C' )
              candidateHighScore.name[hscorelen-1] = '2';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
          case '3' :
            if( candidateHighScore.name[hscorelen-1] == '3' )
              candidateHighScore.name[hscorelen-1] = 'D';
            else if( candidateHighScore.name[hscorelen-1] == 'F' )
              candidateHighScore.name[hscorelen-1] = '3';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
          case '4' :
            if( candidateHighScore.name[hscorelen-1] == '4' )
              candidateHighScore.name[hscorelen-1] = 'G';
            else if( candidateHighScore.name[hscorelen-1] == 'I' )
              candidateHighScore.name[hscorelen-1] = '4';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
          case '5' :
            if( candidateHighScore.name[hscorelen-1] == '5' )
              candidateHighScore.name[hscorelen-1] = 'J';
            else if( candidateHighScore.name[hscorelen-1] == 'L' )
              candidateHighScore.name[hscorelen-1] = '5';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
          case '6' :
            if( candidateHighScore.name[hscorelen-1] == '6' )
              candidateHighScore.name[hscorelen-1] = 'M';
            else if( candidateHighScore.name[hscorelen-1] == 'O' )
              candidateHighScore.name[hscorelen-1] = '6';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
          case '7' :
            if( candidateHighScore.name[hscorelen-1] == '7' )
              candidateHighScore.name[hscorelen-1] = 'P';
            else if( candidateHighScore.name[hscorelen-1] == 'S' )
              candidateHighScore.name[hscorelen-1] = '7';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
          case '8' :
            if( candidateHighScore.name[hscorelen-1] == '8' )
              candidateHighScore.name[hscorelen-1] = 'T';
            else if( candidateHighScore.name[hscorelen-1] == 'V' )
              candidateHighScore.name[hscorelen-1] = '8';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
          case '9' :
            if( candidateHighScore.name[hscorelen-1] == '9' )
              candidateHighScore.name[hscorelen-1] = 'W';
            else if( candidateHighScore.name[hscorelen-1] == 'Z' )
              candidateHighScore.name[hscorelen-1] = '9';
            else
              candidateHighScore.name[hscorelen-1]++;
            break;
        }        
        ticks_last_key = 0;
      }
    switch(k){
      case EStdKeyBackspace :
      case EStdKeyLeftArrow :
#ifdef UIQ3
      case EStdDeviceKeyFourWayLeft:
#endif
        if( hscorelen > 0 )
          candidateHighScore.name[hscorelen-1] = '\0';
        break;
#ifdef UIQ3
	  case EStdDeviceKeyDone:
	  case EStdDeviceKeyAction:
      case EStdKeyEnter:
	  // Special support for jogdial in for SE phones
      case EStdKeyDevice1B:
#else
      case EStdKeyDevice3 :
#endif
        // Store highscore
        if( hscorelen!=0 ){
          if(highscores.Count()<HIGHSCORELIMIT)
            highscores.Append(candidateHighScore);
          else
            highscores[HIGHSCORELIMIT-1] = candidateHighScore;
          THighScoreRecordLinearOrder sortorder(HighScoreSort);
          highscores.Sort(sortorder);
          editEnded = ETrue;
          save_hiscores();
 
          TUint8 *cGreatGame = (TUint8*)"GREAT GAME!";
          TInt xpos = (iGameRes.iWidth-fntNormal->GetTextLength(cGreatGame))/2;
          TInt ypos = (iGameRes.iHeight+szVoidPanel.iHeight)/2 - 35;
          fntNormal->Print(xpos, ypos, cGreatGame);
          play_sound(smpMenu_selected);
          Actions->BitBltBack(*imgOrigBackground);
          Actions->FlipBackBuffer();
          ticks_last_key = 0;
        } else {
          play_sound(smpCancel);
          menu();
          return;
        }
        break;
      default: 
		  {
			   if( (k >= 'A') && (k <= 'Z') &&   hscorelen < HIGHSCORELIMIT ) {
					 candidateHighScore.name[hscorelen] = k;
				}
		  }break;
    }
    
    TInt x_echo = (iGameRes.iWidth-szVoidPanel.iWidth)/2 + 44;
    Actions->BitBlt(BLTTYPE_NORMAL, *imgOrigBackground, ptZero);
    fntNormal->Print(x_echo, hs_ypos, (TUint8*)candidateHighScore.name);
    Actions->FlipBackBuffer();
    sccPreviousKeyEvent = k;
  }
}

void CFrozenBubbleMain::ask_keys(void) {
//  sccLastKeyEvent = EStdKeyNull;
  TSize szVoidPanel = imgVoidPanel->GetBitmap()->SizeInPixels();
  TInt xpos_panel = (iGameRes.iWidth-szVoidPanel.iWidth)/2;
  TInt ypos_panel = (iGameRes.iHeight-szVoidPanel.iHeight)/2;
  Actions->BitBlt(BLTTYPE_MASKED, *imgVoidPanel, TPoint(xpos_panel, ypos_panel));

  hs_ypos = ypos_panel + 5;
  TUint8 *cPleaseEnterKeys = (TUint8*)"PLEASE ENTER KEYS";
  fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cPleaseEnterKeys))/2, hs_ypos, cPleaseEnterKeys);
  hs_ypos += 22;
  hs_ypos += 5;
  
  SET_left = SET_right = SET_fire = SET_center = EStdKeyNull;

  TUint8 *cLEFT = (TUint8*)"LEFT?";
  fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cLEFT))/2, hs_ypos, cLEFT);
  hs_ypos += 22;
  Actions->FlipBackBuffer();

  editEnded = EFalse;
  
  setGameLoop(ask_keysloop_);
}

TInt CFrozenBubbleMain::ask_keysloop_(TAny* aPtr){
  CFrozenBubbleMain *iMain = (CFrozenBubbleMain*)aPtr;
  iMain->ask_keysloop();
  return 0;
}

void CFrozenBubbleMain::ask_keysloop(void){
  TStdScanCode k = sccLastKeyEvent;
  if(editEnded){
    ticks_last_key++;
    if(ticks_last_key > 60)
      menu();
    return;
  }
  if( (k != EStdKeyNull) && (!editEnded) ){
    if( SET_left == EStdKeyNull ){
      SET_left = k;
      TUint8 *cRIGHT = (TUint8*)"RIGHT?";
      fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cRIGHT))/2, hs_ypos, cRIGHT);
      hs_ypos += 22;
    } else if( SET_right == EStdKeyNull ){
      SET_right = k;
      TUint8 *cFIRE = (TUint8*)"FIRE?";
      fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cFIRE))/2, hs_ypos, cFIRE);
      hs_ypos += 22;
    } else if( SET_fire == EStdKeyNull ){
      SET_fire = k;
      TUint8 *cCENTER = (TUint8*)"CENTER?";
      fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cCENTER))/2, hs_ypos, cCENTER);
      hs_ypos += 22;
    } else {
      SET_center = k;
      TUint8 *cTHANKS = (TUint8*)"THANKS!";
      fntNormal->Print((iGameRes.iWidth-fntNormal->GetTextLength(cTHANKS))/2, hs_ypos, cTHANKS);
      hs_ypos += 22;
      editEnded = ETrue;
      KEYS_left   = SET_left;
      KEYS_right  = SET_right;
      KEYS_fire   = SET_fire;
      KEYS_center = SET_center;

      save_settings();
    }

    play_sound(smpTypewriter);

    Actions->FlipBackBuffer();
    ticks_last_key = 0;
    sccLastKeyEvent = EStdKeyNull;
  }
}

void CFrozenBubbleMain::display_highscores(void) {
//  sccLastKeyEvent = EStdKeyNull;
  Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_hiscores, ptZero);

  TBool display_on_app_disabled = ETrue;
  POStop_limit = POSinit_top_limit;

  TInt initial_high_posx = 26;
  TInt high_posx = initial_high_posx;
  TInt high_posy = 36;

  for(TInt idx=0; idx<highscores.Count(); idx++){
    fntDark->Print(high_posx, high_posy, (TUint8*)highscores[idx].name);
    high_posy += 22;
    if(high_posy > 300)
      break;
  }
  Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_hiscores, TPoint(118, 0), TRect(TPoint(118, 0), TSize(120, 320)));    
  high_posy = 36;
  for(TInt idx=0; idx<highscores.Count(); idx++){
    TBuf8<8> bLevel( _L8("L-") );

    bLevel.AppendNum( highscores[idx].level );

    const TUint8* cLevel = bLevel.PtrZ();
    fntDark->Print(high_posx+94, high_posy, cLevel);

    TInt min = highscores[idx].seconds / 60;
    TInt sec = highscores[idx].seconds % 60;
    TBuf8<12> bTime(3);
    bTime.Num(min);
    bTime.Append('\'');
    bTime.AppendNumFixedWidth(sec, EDecimal, 2);
    bTime.Append( _L8("''") );

    fntDark->Print(high_posx+150, high_posy, bTime.PtrZ() );
    high_posy += 22;
    if(high_posy > 300)
      break;
  }
  Actions->FlipBackBuffer();
  display_on_app_disabled = EFalse;

  setGameLoop(highscoreloop_);
}

void CFrozenBubbleMain::highscoreloop(void){
  if( sccLastKeyEvent != EStdKeyNull ){
    sccLastKeyEvent = EStdKeyNull;
    setGameLoop(menu_loop_);
  }
}

TInt CFrozenBubbleMain::highscoreloop_(TAny* aPtr){
  CFrozenBubbleMain *iMain = (CFrozenBubbleMain*)aPtr;
  iMain->highscoreloop();
  return 0;
}

TBool CFrozenBubbleMain::checksaved(void){
  if( Actions->CheckIniFile(uidSavedGame) ){
    Actions->BitBltBack(*imgOrigBackground);
    AddBrightness(*imgOrigBackground);
    AddBrightness(*imgOrigBackground);
    Actions->BitBlt(BLTTYPE_NORMAL, *imgOrigBackground, ptZero);  
    fntDark->Print(12, 146, (TUint8*)"CONTINUE SAVED GAME?");
    fntDark->Print(2, 294, (TUint8*)"YES");
    fntDark->Print(212, 294, (TUint8*)"NO");
    Actions->FlipBackBuffer();
    setGameLoop(NULL);
    global_state = GS_LOADGAME;
    return ETrue;
  } else
    return EFalse;
}

void CFrozenBubbleMain::querysavegame(void){
  Actions->BitBltBack(*imgOrigBackground);
  AddBrightness(*imgOrigBackground);
  AddBrightness(*imgOrigBackground);
  Actions->BitBlt(BLTTYPE_NORMAL, *imgOrigBackground, ptZero);  
  fntDark->Print(20,  85, (TUint8*)"SAVE CURRENT GAME?");
  fntDark->Print(2,  115, (TUint8*)"IF YOU DO NOT SAVE YOUR");
  fntDark->Print(10, 137, (TUint8*)"GAME YOU WILL HAVE TO");
  fntDark->Print(48, 159, (TUint8*)"START FROM THE");
	fntDark->Print(68, 181, (TUint8*)"BEGINNING");
  fntDark->Print(2, 294, (TUint8*)"YES");
  fntDark->Print(212, 294, (TUint8*)"NO");
  Actions->FlipBackBuffer();
  setGameLoop(NULL);
  global_state = GS_SAVEGAME;
}

void CFrozenBubbleMain::dopause(void){
  Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_paused, ptZero);
  pauseIndex = 0;
  pauseSide = 1;
  smxMain->PauseMusic(ETrue);

  setGameLoop(dopause_loop_);
}

TInt CFrozenBubbleMain::dopause_loop_(TAny* aPtr){
  CFrozenBubbleMain *iMain = (CFrozenBubbleMain*)aPtr;
  iMain->dopause_loop();
  return 0;
}

void CFrozenBubbleMain::dopause_loop(void){  
  Actions->BitBlt(BLTTYPE_NORMAL, *imgPause[pauseIndex], TPoint(POSpause_clip.iX, POSpause_clip.iY));
  Actions->FlipBackBuffer();
  if(frand() < 0.05){
    if(frand() < 0.2)
      play_sound(smpSnore);
    pauseIndex = (TInt8)(pauseIndex + pauseSide);
    if( pauseIndex == sizeof(imgPause)/sizeof(*imgPause) ) {
      pauseSide = -1;
      pauseIndex -= 2;
    }
    if( pauseIndex == -1 ){
      pauseSide = 1;
      pauseIndex = 1;
    }
  }
  if( sccLastKeyEvent != EStdKeyNull ){
    ExternalEvent(EVT_ESCAPE);
  }
}

TInt CFrozenBubbleMain::menu_loop_(TAny* aPtr){
  CFrozenBubbleMain *iMain = (CFrozenBubbleMain*)aPtr;
  iMain->menu_loop();
  return 0;
}

void CFrozenBubbleMain::menu_loop(void){
  User::ResetInactivityTime();
  
  Actions->BitBlt(BLTTYPE_NORMAL, *imgBackstartfull, ptZero);

  const TInt MENU_XPOS = 40, MENU_FIRSTY = 72, SPACING = 48;

  switch( sccLastKeyEvent ){
    case EStdKeyRightArrow:
    case EStdKeyDownArrow:
#ifdef UIQ3
    case EStdDeviceKeyFourWayRight:
    case EStdDeviceKeyFourWayDown:
	case EStdDeviceKeyTwoWayDown:
#endif
      if( iMenuCurrent_pos < 3 ){
        iMenuCurrent_pos++;
        play_sound(smpMenu_change);
      }
      break;
#ifdef UIQ3
    case EStdDeviceKeyFourWayLeft:
    case EStdDeviceKeyFourWayUp:
	case EStdDeviceKeyTwoWayUp:
#endif
    case EStdKeyLeftArrow:
    case EStdKeyUpArrow:
      if( iMenuCurrent_pos > 0 ){
        iMenuCurrent_pos--;
        play_sound(smpMenu_change);
      }
      break;
#ifdef UIQ3
     case EStdDeviceKeyAction:
     case EStdDeviceKeyDone:
	 case EStdKeyEnter:
	 // Special support for jogdial in for SE phones
     case EStdKeyDevice1B:
#else
      case EStdKeyDevice3 :
#endif
      play_sound(smpMenu_selected);
      switch(iMenuCurrent_pos){
        case 0:
          bLoadedGame = EFalse;
          LevelsCurrent = 1;
          InitTimer();
          if(!checksaved())
            new_game(EFalse);
          return;
        case 1: 
          bSoundEnabled = !bSoundEnabled;
          save_settings();
          if(bSoundEnabled)
            smxMain->Resume();
//            smxMain->StartMixer();
          else
            smxMain->Pause();
//            smxMain->StopMixer();
          break;
        case 2:
          ask_keys();
          return;
        case 3:
          display_highscores();
          return;
      }
      break;
    default:
      break;
  }
  sccLastKeyEvent = EStdKeyNull;
  for(TUint idxBanner=0; idxBanner<sizeof(BANNERS)/sizeof(*BANNERS); idxBanner++){
    TInt xpos = BANNERS[idxBanner] - iBannerPos;
    CExtendedBitmap *imgBanner;
    switch(idxBanner){
      case 1: imgBanner = imgBannerArtwork;
        break;
      case 2: imgBanner = imgBannerSoundtrack;
        break;
	  case 3: imgBanner = imgBannerCpucontrol;
        break;
      default:imgBanner = imgBannerPorting;
    }
    TSize szBanner = imgBanner->GetBitmap()->SizeInPixels();

    if(xpos > BANNERS_MAX/2)
      xpos = BANNERS[idxBanner] - (iBannerPos + BANNERS_MAX);

    if( (xpos < BANNER_MAXX) && (xpos + szBanner.iWidth >= 0) ){
      TPoint pt(BANNER_MINX, BANNER_Y);
      TRect rct( TPoint(-xpos, 0), TSize(Min(szBanner.iWidth+xpos, BANNER_MAXX-BANNER_MINX), szBanner.iHeight) );
      Actions->BitBlt(BLTTYPE_MASKED, *imgBanner, pt, rct);
    }
  }
  iBannerPos++;
  if(iBannerPos >= BANNERS_MAX)
    iBannerPos = 1;

  // menu screen update
  TPoint iMenuPoint(MENU_XPOS, MENU_FIRSTY);
  CExtendedBitmap *bmpMenuChoice;
  bmpMenuChoice = iMenuCurrent_pos == 0 ? imgTxt_1pgame_over : imgTxt_1pgame_off;
  Actions->BitBlt(BLTTYPE_MASKED, *bmpMenuChoice, iMenuPoint);
  iMenuPoint.iY += SPACING;
  bmpMenuChoice = iMenuCurrent_pos == 1 ? (bSoundEnabled ? imgTxt_sound_act_over : imgTxt_sound_over) : (bSoundEnabled ? imgTxt_sound_act_off : imgTxt_sound_off);
  Actions->BitBlt(BLTTYPE_MASKED, *bmpMenuChoice, iMenuPoint);
  iMenuPoint.iY += SPACING;
  bmpMenuChoice = iMenuCurrent_pos == 2 ? imgTxt_keys_over : imgTxt_keys_off;
  Actions->BitBlt(BLTTYPE_MASKED, *bmpMenuChoice, iMenuPoint);
  iMenuPoint.iY += SPACING;
  bmpMenuChoice = iMenuCurrent_pos == 3 ? imgTxt_highscores_over : imgTxt_highscores_off;
  Actions->BitBlt(BLTTYPE_MASKED, *bmpMenuChoice, iMenuPoint);
  //fix button overlapping
//  Actions->BitBlt(BLTTYPE_NORMAL, *imgBackstartfull, TPoint(135,192), TRect(135,192,176,208) );
  Actions->FlipBackBuffer();
}

void CFrozenBubbleMain::new_game(TBool loadgame) {
  // Check if finished all levels
  if( LevelsCurrent == 101 ){
    if( bLoadedGame )
      Actions->RemoveIniData(uidSavedGame);
    menu();
    return;
  }
  play_music(_L8("frozen-mainzik-1p.it"));
  pdataScore = LevelsCurrent;

  Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_1p, ptZero);

  actionsLeft = actionsRight = actionsFire = actionsCenter = EFalse;
  angle = PI/2;
  
  sticked_bubbles.Reset();
  falling_bubble.Reset();
  exploding_bubble.Reset();
  chains_falling_chained.Reset();
  chains_chained_bubbles.Reset();

  launched_bubble.iColorIndex = BUBBLE_NULL;
  sticking_bubble.iColorIndex = BUBBLE_NULL;
  pdataNewroot = 0;
  pdataNewroot_prelight = 0;
  pdataOddswap = 0;
  pdataHurry = 0;
  pdataNewrootlevel = 0;
  pdataPing_rightImg = 0;
  pdataPing_rightState = STATE_NORMAL;

  print_scores();

  print_compressor();
  if(LevelsCurrent){
    open_level(LevelsCurrent);
  } else {
    for(TInt cy=0; cy<=4; cy++){
      for(TInt cx=0; cx<=(6 + even(cy)); cx++) {
        TBubble b = create_bubble(EFalse);
        real_stick_bubble(b, cx, cy, EFalse);  // this doesn't map well to the 'iter_players' subroutine..
      }
    }
  }
  
  next_bubble = create_bubble(ETrue);
  generate_new_bubble(BUBBLE_NULL);
  Actions->BitBltBack( *imgOrigBackground );
  pdataState = STATE_GAME;

  if( loadgame ){
    load_game();
    bLoadedGame = ETrue;
  }

  RebuildOriginalBackground();
  
  setGameLoop(maingame_);
}

TInt CFrozenBubbleMain::maingame_(TAny* aPtr){
  CFrozenBubbleMain *iMain = (CFrozenBubbleMain*)aPtr;
  iMain->maingame();
  return 0;
}

// ----------- mainloop ---------------------------------------------------

void CFrozenBubbleMain::maingame(void) {
// -------------
//    sccLastKeyEvent = KEYS_fire;
// -------------
  User::ResetInactivityTime();
  if(bFirstLoopIteration){
    // avoid processing the key events on first iteration
    sccLastKeyEvent = EStdKeyNull;
    sccPressedButton = EStdKeyNull;
    bFirstLoopIteration = EFalse;
  }
  
  handle_graphics(EFalse);
  update_game();
  handle_graphics(ETrue);

  Actions->FlipBackBuffer();
}

//- ----------- game stuff -------------------------------------------------

void CFrozenBubbleMain::handle_graphics(TBool iPutImage) {
  // iPutImage = ETrue => use 'put_image'
  // iPutImage = EFalse => use 'erase_image'

  //- bubbles
  if(launched_bubble.iColorIndex != BUBBLE_NULL)
    if(iPutImage)
      Actions->BitBlt(BLTTYPE_MASKED, *imgBall[launched_bubble.iColorIndex], TPoint((TInt)launched_bubble.x, (TInt)launched_bubble.y) );
    else
      erase_image( *imgBall[launched_bubble.iColorIndex], TPoint((TInt)launched_bubble.x, (TInt)launched_bubble.y) );
  if(iPutImage && (tobe_launched.iColorIndex != BUBBLE_NULL))
    Actions->BitBlt(BLTTYPE_MASKED, *imgBall[tobe_launched.iColorIndex], TPoint((TInt)tobe_launched.x, (TInt)tobe_launched.y) );

//-----------
/*  if(iPutImage && pdataNewroot_prelight) {
    if (pdataNewroot_prelight_step++ > 30*pdataNewroot_prelight) {
      pdataNewroot_prelight_step = 0;
    }
    if (pdataNewroot_prelight_step <= 8) {
      TBool hurry_overwritten = EFalse;
      for(TInt idx=0; idx<sticked_bubbles->Count(); idx++){
        TBubble &b = (*sticked_bubbles)[idx];
        if(b.cx+1 == pdataNewroot_prelight_step)
          Actions->BitBlt(BLTTYPE_MASKED, *imgBall[b.iColorIndex], TPoint((TInt)b.x, (TInt)b.y) );
//          put_image(b->{img}, $b->{'x'}, $b->{'y'});
        if(b.cx == pdataNewroot_prelight_step)
          Actions->BitBlt(BLTTYPE_MASKED, *imgBubblesAnimWhite, TPoint((TInt)b.x, (TInt)b.y) );
//          put_image($bubbles_anim{white}, $b->{'x'}, $b->{'y'});
        if(b.cy > 6)
          hurry_overwritten = ETrue;
      }
//      if(hurry_overwritten && pdataHurry_save_img)
//        print_hurry(1);  //- hurry was potentially overwritten
    }
  }*/
//---------- 

  if (sticking_bubble.iColorIndex != BUBBLE_NULL) {
    TBubble &b = sticking_bubble;
    if(iPutImage){
      if (pdataSticking_step == sizeof(imgBubblesAnim)/sizeof(*imgBubblesAnim)) {
        sticking_bubble.iColorIndex = BUBBLE_NULL;
      } else {
        Actions->BitBlt(BLTTYPE_MASKED, *imgBubblesAnim[pdataSticking_step], TPoint((TInt)b.x, (TInt)b.y) );
        if (pdataSticking_step_slowdown) {
          pdataSticking_step++;
          pdataSticking_step_slowdown = 0;
        } else {
          pdataSticking_step_slowdown++;
        }
      }
    } else {
      Actions->BitBlt(BLTTYPE_MASKED, *imgBall[b.iColorIndex], TPoint((TInt)b.x, (TInt)b.y) );
    }
  }
  //- shooter
  TInt num = (TInt)(angle*CANON_ROTATIONS_NB/(PI/2) + 0.5);//-CANON_ROTATIONS_NB;
  TPoint ptCanon((POSleft_limit+POSright_limit)/2 - 34 + canondata->data[num][0], POSinitial_bubble_y + 11 - 34 + canondata->data[num][1]);
  if(iPutImage)
    Actions->BitBlt(BLTTYPE_MASKED, *imgCanon[num], ptCanon ); //- 50/50 stand for half width/height of gfx/shoot/base.png
  else
    erase_image(*imgCanon[num], ptCanon );

  TPoint ptPenguin(POSleft_limit+POSpinguin.iX, POSpinguin.iY);
  CExtendedBitmap &bmpTempPinguin = *StateToPinguinImage(pdataPing_rightState, pdataPing_rightImg);//*imgPinguinNormal[pdataPing_rightImg];
  if(iPutImage)
    Actions->BitBlt(BLTTYPE_MASKED, bmpTempPinguin, ptPenguin );
  else
    erase_image(bmpTempPinguin, ptPenguin );

  //- moving bubbles --> I want them on top of the rest
  if(iPutImage){
//    put_allimages_to_background(malus_bubble);
    put_allimages_to_background(falling_bubble);
    put_allimages_to_background(exploding_bubble);
  }else{
//    remove_images_from_background(malus_bubble);
    remove_images_from_background(falling_bubble);
    remove_images_from_background(exploding_bubble);
  }
}

//- ----------- mainloop helper --------------------------------------------

void CFrozenBubbleMain::update_game(void) {
  if(pdataState == STATE_GAME){
    handle_game_events();
    if (actionsLeft)  angle += LAUNCHER_SPEED;
    if (actionsRight) angle -= LAUNCHER_SPEED;
    if (actionsCenter) {
      if (angle >= PI/2 - LAUNCHER_SPEED && angle <= PI/2 + LAUNCHER_SPEED) {
        angle = PI/2;
      } else {
        angle += (angle < PI/2) ? LAUNCHER_SPEED : -LAUNCHER_SPEED;
      }
    }
    if(angle < 0.1) angle = 0.1;
    if(angle > PI-0.1) angle = PI-0.1;
    pdataHurry++;
    
    if (pdataHurry > TIME_HURRY_WARN) {
      TInt oddness = odd(TInt((pdataHurry-TIME_HURRY_WARN)/(500/TARGET_ANIM_SPEED))+1);
      if (log_xor(pdataHurry_oddness, oddness)) { 
        if (oddness) {
          play_sound(smpHurry);
          print_hurry();
        } else {
          remove_hurry();
        }
      }

      pdataHurry_oddness = oddness;
    }
    
    if ((actionsFire || pdataHurry == TIME_HURRY_MAX) && (launched_bubble.iColorIndex==BUBBLE_NULL) && !(AnyExistsChaindestx(falling_bubble)) /*&& (malus_bubble->Count()==0)*/) {
      play_sound(smpLaunch);
      launched_bubble = tobe_launched;
      launched_bubble.direction = angle;
      tobe_launched.iColorIndex = BUBBLE_NULL;
      actionsFire = 0;
      actionsHadfire = 1;
      pdataHurry = 0;

      remove_hurry();

    }
    if(launched_bubble.iColorIndex != BUBBLE_NULL){
      launched_bubble.x_old = launched_bubble.x; // save coordinates for potential collision
      launched_bubble.y_old = launched_bubble.y;
      launched_bubble.x += BUBBLE_SPEED * cos(launched_bubble.direction);
      launched_bubble.y -= BUBBLE_SPEED * sin(launched_bubble.direction);
      if (launched_bubble.x < POSleft_limit) {
        play_sound(smpRebound);
        launched_bubble.x = 2 * POSleft_limit - launched_bubble.x;
        launched_bubble.direction -= 2*(launched_bubble.direction-PI/2);
      }
      if (launched_bubble.x > POSright_limit - BUBBLE_SIZE) {
        play_sound(smpRebound);
        launched_bubble.x = 2 * (POSright_limit - BUBBLE_SIZE) - launched_bubble.x;
        launched_bubble.direction += 2*(PI/2-launched_bubble.direction);
      }

      if (launched_bubble.y <= POStop_limit + pdataNewrootlevel * ROW_SIZE) {
        TPoint cpt = get_array_closest_pos((TInt)launched_bubble.x, (TInt)launched_bubble.y);
        stick_bubble(launched_bubble, cpt.iX, cpt.iY, 1);
        launched_bubble.iColorIndex = BUBBLE_NULL;
      } else {
        for(TInt idx=0; idx<sticked_bubbles.Count(); idx++){
          TBubble &cbubble = sticked_bubbles[idx];
          if(is_collision(launched_bubble, cbubble.x, cbubble.y)){
            TPoint cpt = get_array_closest_pos((TInt)(launched_bubble.x_old+launched_bubble.x)/2, (TInt)(launched_bubble.y_old+launched_bubble.y)/2);
            stick_bubble(launched_bubble, cpt.iX, cpt.iY, 1);
            launched_bubble.iColorIndex = BUBBLE_NULL;
            break;
          }
        }
      }
    }

    if(tobe_launched.iColorIndex == BUBBLE_NULL)
      generate_new_bubble(BUBBLE_NULL);

    if ((!actionsLeft) && (!actionsRight) && (!actionsHadfire)) {
      pdataSleeping++;
    } else {
      pdataSleeping = 0;
      pdataPing_rightMovelatency = -20;
    }
    if (pdataSleeping > TIMEOUT_PINGUIN_SLEEP) {
      pdataPing_rightState = STATE_SLEEP;
    } else if (pdataPing_rightState == STATE_SLEEP) {
      pdataPing_rightState = STATE_NORMAL;
    }
    if (pdataPing_rightState == STATE_RIGHT && !(actionsRight)
     || pdataPing_rightState == STATE_LEFT && !(actionsLeft)
     || pdataPing_rightState == STATE_ACTION && (pdataPing_rightActionlatency++ > 5)) {
      pdataPing_rightState = STATE_NORMAL;
    }
    if(actionsRight)
      pdataPing_rightState = STATE_RIGHT;
    if(actionsLeft)
      pdataPing_rightState = STATE_LEFT;
    if (actionsHadfire) {
      pdataPing_rightState = STATE_ACTION;
      actionsHadfire = 0;
      pdataPing_rightActionlatency = 0;
    }

    if (pdataPing_rightState == STATE_NORMAL && (pdataPing_rightMovelatency++ > 10)) {
      pdataPing_rightMovelatency = 0;
      if(frand() < 0.4)
        pdataPing_rightImg = rand(sizeof(imgPinguinNormal)/sizeof(*imgPinguinNormal));
    }

    if(pdataPing_rightImg >= (TInt)(STATE_NORMAL==pdataPing_rightState ? sizeof(imgPinguinNormal)/sizeof(*imgPinguinNormal) : 1) )
      pdataPing_rightImg = 0;

    verify_if_end();
  } else if(pdataState == STATE_LOST){

    if (sticked_bubbles.Count() > 0) {
      TBubble b = sticked_bubbles[ sticked_bubbles.Count()-1 ] ;
      sticked_bubbles.Remove( sticked_bubbles.Count()-1 );
      
      Actions->BitBlt(BLTTYPE_MASKED, *imgBubblesAnimLose, TPoint( (TInt)(b.x), (TInt)(b.y) ) );
      if (sticked_bubbles.Count() == 0) {
        Actions->BitBlt(BLTTYPE_MASKED, *imgLose, TPoint(POScenterpanel.iX, POScenterpanel.iY) );
        play_sound(smpNoh);
      }

    } else {
      if( sccLastKeyEvent!=EStdKeyNull )
        new_game(EFalse);
    }
  } else if(pdataState == STATE_WON){
    if (exploding_bubble.Count() == 0) {
      if( sccLastKeyEvent!=EStdKeyNull )
        new_game(EFalse);
    }
  }

  TInt idx=0; 
  tmp_bubbles.Reset();
  for(idx=0; idx<falling_bubble.Count(); idx++){
    TBubble &b = falling_bubble[idx];
    if (b.wait_fall) {
      b.wait_fall--;
    } else {
      if ( (b.chaindestx != -1) && (b.y > 175 || b.chaingoingup)) {
        TReal acceleration = FREE_FALL_CONSTANT*3;
        if (!b.chaingoingup) {
          TReal time_to_zero = b.speed/acceleration;
          TReal distance_to_zero = b.speed * (b.speed/acceleration + 1) / 2;
          TReal time_to_destination = (-1 + sqrt(1 + 8/acceleration*(b.y-b.chaindesty+distance_to_zero))) / 2;
          b.speedx = (b.chaindestx - b.x) / (time_to_zero + time_to_destination);
          b.chaingoingup = 1;
        }
        b.speed -= acceleration;
        b.x += b.speedx;
        if (abs(b.x - b.chaindestx) < abs(b.speedx)) {
          b.x = b.chaindestx;
          b.speedx = 0;
        }
        b.y += b.speed;
        if(b.y < b.chaindesty)
          tmp_bubbles.Append(b);
      } else {
        b.y += b.speed;
        b.speed += FREE_FALL_CONSTANT;
        if( b.y > iGameRes.iHeight )
          falling_bubble.Remove(idx);
      }
    }
    if((b.y > 200) && (b.chaindestx==-1)) 
      tmp_bubbles.Append(b);
  }
  for(idx=0; idx<tmp_bubbles.Count(); idx++){
    TBubble &b = tmp_bubbles[idx];
    if( b.chaindestx != -1 ){
      // remove bubble from chains_falling_chained
      TInt idxChained = FindBubbleInArray(b, chains_falling_chained);
      if( idxChained != -1 )
        chains_falling_chained.Remove( idxChained );
      // remove bubble from chains_chained_bubbles
      idxChained = FindBubbleInArray(b, chains_chained_bubbles);
      if( idxChained != -1 )
        chains_chained_bubbles.Remove( idxChained );
      stick_bubble(b, b.chaindestcx, b.chaindestcy, ETrue);
    }
  }

  idx=0;
  while( idx<exploding_bubble.Count() ){
    TBubble &b = exploding_bubble[idx];
    b.x += b.speedx;
    b.y += b.speedy;
    b.speedy += FREE_FALL_CONSTANT;
    if( b.y > iGameRes.iHeight )
      exploding_bubble.Remove(idx);
    else
      idx++;
  }

  if( ((pdataPing_rightState == STATE_LOST) || (pdataPing_rightState == STATE_WON)) && (pdataPing_rightMovelatency++ > 5) ){
    pdataPing_rightMovelatency = 0;
    pdataPing_rightImg++;
    if( pdataPing_rightImg == 9 ) //fixed just for simplicity
      pdataPing_rightImg = (pdataPing_rightState == STATE_WON) ? pinguinWin_roll_back_index : pinguinLose_roll_back_index;//  $pinguin{"$state".'_roll_back_index'};
  }
}

void CFrozenBubbleMain::stick_bubble(TBubble &bubble, TInt xpos, TInt ypos, TBool count_for_root) {
  RBubbleArray falling;// = new(ELeave) RBubbleArray;
//  CleanupStack::PushL( falling );
  CleanupClosePushL(falling);
  
  TInt idx;
  for(idx=0; idx<sticked_bubbles.Count(); idx++){
    TBubble &StickedBubble = sticked_bubbles[idx];
    if( bubble_next_to(StickedBubble.cx, StickedBubble.cy, xpos, ypos) )
      bubble.addNeighbour( StickedBubble );
  }
  bubble.cx = (TInt8)xpos;
  bubble.cy = (TInt8)ypos;

  find_bubble_group(bubble, will_destroy);
  // remove bubble
  idx = FindBubbleInArray(bubble, will_destroy);
  if(idx!=-1)
    will_destroy.Remove(idx);
  if (will_destroy.Count() <= 1) {
    // stick
    play_sound(smpStick);
    real_stick_bubble(bubble, xpos, ypos, ETrue);
    sticking_bubble = bubble;
    pdataSticking_step = 0;
  } else {
    //- destroy the group
    play_sound(smpDestroy_group);

    // remove will_destroy bubbles from sticked_bubble neighbours
    for(idx=0; idx<sticked_bubbles.Count(); idx++){
      TBubble &sbubble = sticked_bubbles[idx];
      for(TInt idxNB=0; idxNB<sbubble.neighboursCount();){
        TInt idxNeighbour = FindBubbleInArray(sbubble.neighbours[idxNB], will_destroy);
        if(idxNeighbour != -1)
          sbubble.removeNeighbour(idxNB);
        else
          idxNB++;
      }
    }

    difference2(sticked_bubbles, will_destroy, sticked_bubbles);

    calc_real_pos(bubble);
    destroy_bubbles(will_destroy);
    destroy_bubbles(bubble);
  
    //- find falling bubbles
    for(idx=0; idx<sticked_bubbles.Count(); idx++)
      sticked_bubbles[idx].mark = 0;
    RBubbleArray neighbours;// = new(ELeave) RBubbleArray;
    CleanupClosePushL(neighbours);
    
    for(idx=0; idx<sticked_bubbles.Count(); idx++)
      if( sticked_bubbles[idx].cy == pdataNewrootlevel )
        neighbours.Append( sticked_bubbles[idx] );        
    RBubbleArray still_sticked;// = new(ELeave) RBubbleArray;
    CleanupClosePushL(still_sticked);

    TInt8 distance_to_root=0;
    while( neighbours.Count()>0 ) {
      distance_to_root++;
      for(idx=0; idx<sticked_bubbles.Count(); idx++){
        TBubble &sticked_b = sticked_bubbles[idx];
        if( ( sticked_b.mark==0 ) && ( FindBubbleInArray( sticked_b, neighbours )!=-1 ) )
          sticked_b.mark = distance_to_root;
      }
      for(idx=0; idx<neighbours.Count(); idx++)
        still_sticked.Append( neighbours[idx] );
      tmp_bubbles.Reset();
      for(idx=0; idx<neighbours.Count(); idx++){
        TBubble &nb = neighbours[idx];
        for(TInt idxNB=0; idxNB<nb.neighboursCount(); idxNB++){
          TInt idxStickedNB = FindBubbleInArray(nb.neighbours[idxNB], sticked_bubbles);
          if( idxStickedNB != -1 ){
            TBubble &sticked_b = sticked_bubbles[idxStickedNB];
            if( (sticked_b.mark==0) && (FindBubbleInArray(sticked_b, tmp_bubbles)==-1) )
              tmp_bubbles.Append( sticked_bubbles[idxStickedNB] );
          }
        }
      }
      neighbours.Reset();
      for(idx=0; idx<tmp_bubbles.Count(); idx++)
        neighbours.Append( tmp_bubbles[idx] );
    }

    difference2(sticked_bubbles, still_sticked, falling);
    difference2(sticked_bubbles, falling, sticked_bubbles);
    
    CleanupStack::PopAndDestroy(); // To Close() still_sticked
//    CleanupStack::PopAndDestroy(still_sticked); // To destroy

    CleanupStack::PopAndDestroy(); // To Close() neighbours
//    CleanupStack::PopAndDestroy(neighbours); // To destroy

    //- prepare falling bubbles
    TInt max_cy_falling = 0;
    for(idx=0; idx<falling.Count(); idx++)
      if( falling[idx].cy > max_cy_falling ) 
        max_cy_falling = falling[idx].cy;
    TInt shift_on_same_line = 0;
    TInt line = max_cy_falling;
    //- sort bottom-to-up / right-to-left
    TBubbleLinearOrder position_order(BubblePositionOrder);
    CopyBubbleArray(falling, tmp_bubbles);
    tmp_bubbles.Sort( position_order );
    for(idx=0; idx<tmp_bubbles.Count(); idx++){
      TBubble &b = tmp_bubbles[idx];
      if(line != b.cy)
        shift_on_same_line = 0;
      line = b.cy;
      b.wait_fall = (max_cy_falling - b.cy)*5 + shift_on_same_line;
      shift_on_same_line++;
      b.speed = 0;
    }
    AppendBubbleArray(tmp_bubbles, falling_bubble);
    
    remove_images_from_background(will_destroy);
    remove_images_from_background(falling);

    //- redraw neighbours because parts of neighbours have been erased by previous statement
    // find unique neighbours of will_destroy & falling
    tmp_bubbles.Reset();
    AppendNeighboursOf( will_destroy, tmp_bubbles );
    AppendNeighboursOf( falling, tmp_bubbles );
    for(idx=0; idx<tmp_bubbles.Count(); idx++)
      print_bubble( tmp_bubbles[idx] );
//    need_redraw = 1;
  }
  if (count_for_root) {
    pdataNewroot++;
    if (pdataNewroot == TIME_APPEARS_NEW_ROOT-1) {
      pdataNewroot_prelight = 2;
      pdataNewroot_prelight_step = 0;
    }
    if (pdataNewroot == TIME_APPEARS_NEW_ROOT) {
      pdataNewroot_prelight = 1;
      pdataNewroot_prelight_step = 0;
    }
    if (pdataNewroot > TIME_APPEARS_NEW_ROOT) {
//      need_redraw = 1;
      pdataNewroot_prelight = 0;
      play_sound(smpNewroot_solo);
      pdataNewroot = 0;
      pdataOddswap = !pdataOddswap;
      for(idx=0; idx<sticked_bubbles.Count(); idx++){
        TBubble &b = sticked_bubbles[idx];
        b.cy++;
        calc_real_pos(b);
        for(TInt idx2=0; idx2<b.neighboursCount(); idx2++)
          b.neighbours[idx2].cy++;
      }
      for(idx=0; idx<falling_bubble.Count(); idx++){
        TBubble &b = falling_bubble[idx];
        if( b.chaindestx != -1 ){
          b.chaindestcy++;
          b.chaindesty = (TInt16)(b.chaindesty + ROW_SIZE);
        }
      }
      launched_bubble.cy++;
      launched_bubble.y += ROW_SIZE;
      sticking_bubble.cy++;
      sticking_bubble.y += ROW_SIZE;
      calc_real_pos(launched_bubble);
      Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_1p, ptZero );
      put_allimages_to_background(sticked_bubbles);
      pdataNewrootlevel++;
      print_compressor();
    }
  }

/*  if (need_redraw) {
    TInt malus_val = will_destroy.Count() + falling.Count() - 2;
    if(malus_val > 0)
      malus_val += 0;
    if(malus_val < 0)
      malus_val = 0;
  }*/

  CleanupStack::PopAndDestroy(); // To Close() falling
//  CleanupStack::PopAndDestroy(falling); // To destroy
  
  RebuildOriginalBackground();
}

void CFrozenBubbleMain::verify_if_end() {
  for(TInt idx=0; idx<sticked_bubbles.Count(); idx++)
    if( sticked_bubbles[idx].cy > 11) {
      pdataState = STATE_LOST;
      play_sound(smpLose);
      pdataPing_rightState = STATE_LOST;
      pdataPing_rightImg = 0;
      break;
    }

  if (sticked_bubbles.Count() == 0) {
    Actions->BitBlt(BLTTYPE_MASKED, *imgWin, TPoint(POScenterpanel.iX, POScenterpanel.iY) );
    pdataState = STATE_WON;
    pdataPing_rightState = STATE_WON;
    pdataPing_rightImg = 0;
    RebuildOriginalBackground();
    if(LevelsCurrent)
      LevelsCurrent++;
  }
}

void CFrozenBubbleMain::RebuildOriginalBackground(){
  Actions->BitBlt(BLTTYPE_NORMAL, *imgBack_1p, ptZero);
  put_allimages_to_background(sticked_bubbles);
  print_compressor();
  print_scores();
  if(next_bubble.iColorIndex != BUBBLE_NULL)
    print_next_bubble(next_bubble.iColorIndex);
  if(pdataState == STATE_WON)
    Actions->BitBlt(BLTTYPE_MASKED, *imgWin, TPoint(POScenterpanel.iX, POScenterpanel.iY) );
  
  Actions->BitBltBack( *imgOrigBackground );  
}

void CFrozenBubbleMain::destroy_bubbles(RBubbleArray &bubz) {
  for(TInt idx=0; idx<bubz.Count(); idx++)
    destroy_bubbles(bubz[idx]);
}

void CFrozenBubbleMain::destroy_bubbles(TBubble &bub){
  bub.speedx =  frand(3)-1.5;
  bub.speedy = -frand(4)-2;
  exploding_bubble.Append(bub);
}

TBool CFrozenBubbleMain::is_collision(TBubble &bub, TReal x, TReal y) {
  const TReal DISTANCE_COLLISION_SQRED = sqr(BUBBLE_SIZE * 0.82);

  TReal xs = sqr(bub.x - x);
  if(xs > DISTANCE_COLLISION_SQRED)
    return EFalse; 
  return (xs + sqr(bub.y - y)) < DISTANCE_COLLISION_SQRED;
}

void CFrozenBubbleMain::find_bubble_group(TBubble &b, RBubbleArray &group) {
  group.Reset();

  RBubbleArray neighbours;// = new(ELeave) RBubbleArray;
//  CleanupStack::PushL( neighbours );
  CleanupClosePushL( neighbours );

  neighbours.Append(b);
  while( neighbours.Count() != 0 ) {
    TInt idx;
    for(idx=0; idx<neighbours.Count(); idx++)
      group.Append( neighbours[idx] );
    // find neighbours each neighbour
    tmp_bubbles.Reset();
    for(idx=0; idx<neighbours.Count(); idx++){
      TBubble &bNeighbour = neighbours[idx];
      for(TInt idx2=0; idx2<bNeighbour.neighboursCount(); idx2++){
        TBubbleBase &bn = bNeighbour.neighbours[idx2];
        if( (bn.iColorIndex == b.iColorIndex) && (FindBubbleInArray( bn, tmp_bubbles ) == -1) && (FindBubbleInArray( bn, group ) == -1) ){
          TInt idxFounded = FindBubbleInArray( bn, sticked_bubbles );
          if(idxFounded != -1)
            tmp_bubbles.Append( sticked_bubbles[idxFounded] );
        }
      }
    }
    neighbours.Reset();
    for(idx=0; idx<tmp_bubbles.Count(); idx++)
      neighbours.Append( tmp_bubbles[idx] );
  }
  CleanupStack::PopAndDestroy(); // To Close() neighbours
//  CleanupStack::PopAndDestroy(neighbours); // To destroy
}

void CFrozenBubbleMain::handle_game_events(void) {
  // Keydown event
  actionsLeft = actionsRight = actionsFire = actionsCenter = EFalse;
  if(sccPressedButton==KEYS_left)
    actionsLeft   = ETrue;
  else if(sccPressedButton==KEYS_right)
    actionsRight = ETrue;
  else if(sccLastKeyEvent==KEYS_fire)
    actionsFire = ETrue;
  else if(sccPressedButton==KEYS_center)
    actionsCenter = ETrue;
  sccLastKeyEvent = EStdKeyNull;
}

void CFrozenBubbleMain::print_compressor() {
  TInt x = POScompressor_xpos;
  TInt y = POStop_limit + pdataNewrootlevel * ROW_SIZE;

  TSize szimgCompressorExt  = imgCompressorExt->GetBitmap()->SizeInPixels(), 
        szimgCompressorMain = imgCompressorMain->GetBitmap()->SizeInPixels();
  Actions->BitBlt(BLTTYPE_MASKED, *imgCompressorMain, TPoint(x - szimgCompressorMain.iWidth/2, y - szimgCompressorMain.iHeight) );
  

  y -= szimgCompressorMain.iHeight - 3;

  while (y > 0) {
    Actions->BitBlt(BLTTYPE_MASKED, *imgCompressorExt, TPoint(x - szimgCompressorExt.iWidth/2, y - szimgCompressorExt.iHeight) );
    y -= szimgCompressorExt.iHeight;
  }
}

void CFrozenBubbleMain::load_levelset() {
	_LIT( LEVELS, "levels" ); 
	TFileName iLevelsFName(iPrivatePath);
	iLevelsFName.Append(LEVELS);
  RFile file;
  CleanupClosePushL( file );

  User::LeaveIfError( file.Open( EikonEnv->FsSession(), iLevelsFName, EFileStream | EFileRead ) );
  TInt size;
  User::LeaveIfError( file.Size( size ) );

  TUint8* tbuf = new( ELeave )TUint8[ size ];

  TPtr8 ptr( (TUint8*)tbuf, size );
  User::LeaveIfError( file.Read( ptr ) );
  
  CleanupStack::PopAndDestroy( 1 ); // close file
  
  TInt iRow=0, iColumn=0;
  TInt iLevel=0;
  for(TInt idx=0; idx<ptr.MaxLength(); idx++){
    switch(ptr[idx]){
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        levels[iLevel][iRow][iColumn] = (TInt8)(ptr[idx]-'0');
        iColumn++;
        break;
      case '-':
        levels[iLevel][iRow][iColumn] = -1;
        iColumn++;
        break;
      case 13:
        if(iColumn == 0){
          iRow = 0;
          iLevel++;
        } else {
          for(;iColumn<=7;iColumn++){
            levels[iLevel][iRow][iColumn] = -1;
          }
          iRow++;
        }
        iColumn = 0;
        break;
      case ' ':
      default:;
    } 
  }
  delete[] tbuf;
}

void CFrozenBubbleMain::print_scores() {
  TBuf8<8> bLevel(3);
  bLevel.Num(pdataScore);

  const TUint8* pLevel = bLevel.PtrZ();
  fntNormal->Print(POSscoresx-fntNormal->GetTextLength(pLevel)/2, POSscoresy, pLevel);
}

void CFrozenBubbleMain::remove_images_from_background(RBubbleArray &images) {
  for(TInt idx=0; idx<images.Count(); idx++){
    TBubble &bubble = images[idx];
    erase_image( *imgBall[0], TPoint((TInt)bubble.x, (TInt)bubble.y) );
  }
}

void CFrozenBubbleMain::erase_image(CExtendedBitmap &image, const TPoint &ptTarget) {
  TSize szImage = image.GetBitmap()->SizeInPixels();
  Actions->BitBlt(BLTTYPE_NORMAL, *imgOrigBackground, ptTarget, TRect(ptTarget, szImage) );
}

void CFrozenBubbleMain::put_allimages_to_background(RBubbleArray &ba) {
  for(TInt idx=0; idx<ba.Count(); idx++)
    print_bubble( ba[idx] );
}

void CFrozenBubbleMain::difference2(RBubbleArray &barr, RBubbleArray &bdiff, RBubbleArray &bres){
  RBubbleArray btmp;// = new(ELeave) RBubbleArray();
//  CleanupStack::PushL(btmp);
  CleanupClosePushL( btmp );

  TInt idx;
  for(idx=0; idx<barr.Count(); idx++){
    TBool found = EFalse;
    for(TInt idxDF=0; idxDF<bdiff.Count(); idxDF++)
      if( CompareBubbles(barr[idx], bdiff[idxDF]) == 0 ){
        found = ETrue;
        break;
      }
    if(!found)
      btmp.Append(barr[idx]);
  }
  bres.Reset();
  for(idx=0; idx<btmp.Count(); idx++)
    bres.Append( btmp[idx] );
  CleanupStack::PopAndDestroy(); // To Close() btmp
//  CleanupStack::PopAndDestroy(btmp); // To destroy
}

// iLevel = 1..100
void CFrozenBubbleMain::open_level(TInt iLevel) {
  const TInt iAbsoluteLevel=iLevel-1;
  for(TInt iRow=0; iRow<10; iRow++)
    for(TInt iCol=0; iCol<8; iCol++){
        TInt8 color = levels[iAbsoluteLevel][iRow][iCol];
        if(color != BUBBLE_NULL){
          TBubble newbubble = create_bubble_given_img(color);
          real_stick_bubble(newbubble, iCol, iRow, EFalse);
        }
      }
}

// bubble ends its life sticked somewhere
void CFrozenBubbleMain::real_stick_bubble(TBubble &bubble, TInt xpos, TInt ypos, TBool neighbours_ok) {
  bubble.cx = (TInt8)xpos;
  bubble.cy = (TInt8)ypos;

  // this should be done in early stage to calculate real position x, y
  calc_real_pos(bubble);
  
  for(TInt idx=0; idx<sticked_bubbles.Count(); idx++){
    if (bubble_next_to(sticked_bubbles[idx].cx, sticked_bubbles[idx].cy, bubble.cx, bubble.cy)) {
      sticked_bubbles[idx].addNeighbour(bubble);
      if(!neighbours_ok)
        bubble.addNeighbour( sticked_bubbles[idx] );
    }
  }
  sticked_bubbles.Append(bubble);

  Actions->BitBlt(BLTTYPE_MASKED, *imgBall[bubble.iColorIndex], TPoint((TInt)bubble.x, (TInt)bubble.y) );
}

// ----------- bubble game stuff ------------------------------------------

TInt CFrozenBubbleMain::get_array_yclosest(TReal y) {
  // - Corrects the microsoft's rounding issue -
  TInt32 res;
  Math::Int(res, (y-POStop_limit+ROW_SIZE/2) / ROW_SIZE);
  return res;
}

TPoint CFrozenBubbleMain::get_array_closest_pos(TInt x, TInt y) { // roughly the opposite than previous function
  TPoint res;
  res.iY = get_array_yclosest(y);
  res.iX = TInt((x-POSleft_limit+BUBBLE_SIZE/2 - odd(res.iY+pdataOddswap)*BUBBLE_SIZE/2)/BUBBLE_SIZE);
  return res;
}

void CFrozenBubbleMain::calc_real_pos(TBubble &b) {
  calc_real_pos_given_arraypos(b.cx, b.cy, b.x, b.y);
}

void CFrozenBubbleMain::calc_real_pos_given_arraypos(TInt8 cx, TInt8 cy, TReal &newx, TReal &newy) {
  newx = POSleft_limit + cx * BUBBLE_SIZE + odd(cy+pdataOddswap) * BUBBLE_SIZE/2;
  newy = POStop_limit + cy * ROW_SIZE;
}

TBool CFrozenBubbleMain::bubble_next_to(TInt x1, TInt y1, TInt x2, TInt y2) {
  return (sqr(x1+odd(y1+pdataOddswap)*0.5 - (x2+odd(y2+pdataOddswap)*0.5)) + sqr(y1 - y2)) < 3;
}

TBubble CFrozenBubbleMain::create_bubble_given_img(TInt8 iColor) {
  TBubble bubble; 
  bubble.iColorIndex = iColor;
  return bubble;
}

TBubble CFrozenBubbleMain::create_bubble(TBool create_same_color) {
  TBubble b = create_bubble_given_img((TInt8)(rand(sizeof(imgBall)/sizeof(*imgBall))));
//  if(create_same_color && (!IsColorInBubbleArray(b.iColorIndex, *sticked_bubbles)))
//    return create_bubble(ETrue); //- prototype checking pb w/ recursion
  while(create_same_color && (!IsColorInBubbleArray(b.iColorIndex, sticked_bubbles)))
    b = create_bubble_given_img((TInt8)(rand(sizeof(imgBall)/sizeof(*imgBall))));
  return b;    
}

void CFrozenBubbleMain::AppendNeighboursOf(RBubbleArray &bubbles, RBubbleArray &dest){
  for(TInt idx=0; idx<bubbles.Count(); idx++){
    TBubble &nb = bubbles[idx];
    for(TInt idxNB=0; idxNB<nb.neighboursCount(); idxNB++){
      TInt idxSB = FindBubbleInArray( nb.neighbours[idxNB], sticked_bubbles );
      if( (idxSB != -1) && (FindBubbleInArray( nb.neighbours[idxNB], dest ) == -1) )
        dest.Append( sticked_bubbles[idxSB] );
    }
  }
}

TBool CFrozenBubbleMain::IsColorInBubbleArray(TInt8 iColor, RBubbleArray &rarray){
  for(TInt idx=0; idx<rarray.Count(); idx++)
    if(rarray[idx].iColorIndex == iColor)
      return ETrue;
  return EFalse;
}

void CFrozenBubbleMain::print_hurry(/*TBool dont_save_background*/) {
  Actions->BitBlt(BLTTYPE_MASKED, *imgHurry, TPoint(POSleft_limit + POShurry.iX, POShurry.iY) );
}

void CFrozenBubbleMain::remove_hurry() {
  erase_image(*imgHurry, TPoint(POSleft_limit + POShurry.iX, POShurry.iY));
}

void CFrozenBubbleMain::generate_new_bubble(TInt8 iColor) {
  tobe_launched = next_bubble;
  tobe_launched.x = (TInt16)((POSleft_limit+POSright_limit)/2 - BUBBLE_SIZE/2);
  tobe_launched.y = (TInt16)POSinitial_bubble_y;
  next_bubble = iColor != BUBBLE_NULL ? create_bubble_given_img(iColor) : create_bubble(ETrue);
  next_bubble.x = (TInt16)POSnext_bubble.iX; // necessary to keep coordinates, for verify_if_end
  next_bubble.y = (TInt16)POSnext_bubble.iY;
  print_next_bubble(next_bubble.iColorIndex);
  RebuildOriginalBackground();
}

void CFrozenBubbleMain::print_next_bubble(TInt8 iColor) {
  Actions->BitBlt(BLTTYPE_MASKED, *imgBall[iColor], TPoint((TInt)next_bubble.x, (TInt)next_bubble.y) );
}

void CFrozenBubbleMain::print_bubble(TBubble &bubble){
  Actions->BitBlt(BLTTYPE_MASKED, *imgBall[bubble.iColorIndex], TPoint((TInt)bubble.x, (TInt)bubble.y) );
}

TBool CFrozenBubbleMain::AnyExistsChaindestx(RBubbleArray &barray){
  for(TInt idx=0; idx<barray.Count(); idx++)
    if(barray[idx].chaindestx != -1)
      return ETrue;
  return EFalse;
}

TBool CFrozenBubbleMain::AnyChaindestx(RBubbleArray &barray){
  for(TInt idx=0; idx<barray.Count(); idx++)
    if( (barray[idx].chaindestx != -1) && (barray[idx].chaindestx) )
      return ETrue;
  return EFalse;
}

TInt CFrozenBubbleMain::FindBubbleInArray(TBubbleBase &b, RBubbleArray &br){
  TInt idx=0;
  while( (idx<br.Count()) && CompareBubbles(br[idx], b) )
    idx++;
  return idx<br.Count() ? idx : -1;
}

CExtendedBitmap* CFrozenBubbleMain::StateToPinguinImage(TInt state, TInt index){
  switch( state ){
    case STATE_ACTION :
      return imgPinguinAction;
    case STATE_LEFT :
      return imgPinguinMoveLeft;
    case STATE_RIGHT :
      return imgPinguinMoveRight;
    case STATE_SLEEP :
      return imgPinguinSleep;
    case STATE_WON :
      return imgPinguinWin[index];
    case STATE_LOST :
      return imgPinguinLose[index];
    default:
      return imgPinguinNormal[index];
  }
}

TInt CFrozenBubbleMain::CompareBubbles(TBubbleBase &b1, TBubbleBase &b2){
  return (b1.cx != b2.cx) || (b1.cy != b2.cy) || (b1.iColorIndex != b2.iColorIndex);
}

TInt CFrozenBubbleMain::BubblePositionOrder(const TBubble &a, const TBubble &b){
  return b.cy*8 + b.cx -( a.cy*8 + a.cx );
}

void CFrozenBubbleMain::load_settings(void){
  // load ini file settings
  struct TSettingsRecord settings;
  TUint size = sizeof(settings);
  if( Actions->ReadIniFile(uidSettings, &settings, size) ){
    KEYS_left   = settings.keys[0];
    KEYS_right  = settings.keys[1];
    KEYS_fire   = settings.keys[2];
    KEYS_center = settings.keys[3];
    bSoundEnabled = settings.bSound;
  }
}

void CFrozenBubbleMain::save_settings(void){
  // save ini file settings
  struct TSettingsRecord settings;
  settings.keys[0] = KEYS_left;
  settings.keys[1] = KEYS_right;
  settings.keys[2] = KEYS_fire;
  settings.keys[3] = KEYS_center;
  settings.bSound  = bSoundEnabled;
  Actions->WriteIniFile(uidSettings, &settings, sizeof(settings));
}

void CFrozenBubbleMain::load_game(void){
  // save ini file settings
  struct TSavedGameRecord *saved;
  saved = new(ELeave) TSavedGameRecord;
  CleanupStack::PushL( saved );
  TUint ssize = sizeof(*saved);
  if( Actions->ReadIniFile(uidSavedGame, saved, ssize) ){
    pdataScore = saved->pdataScore;
    pdataNewroot = saved->pdataNewroot;
    pdataNewroot_prelight = saved->pdataNewroot_prelight;
    pdataOddswap = saved->pdataOddswap;
    pdataHurry = saved->pdataHurry;
    pdataNewrootlevel = saved->pdataNewrootlevel;
    pdataNewroot_prelight_step = saved->pdataNewroot_prelight_step;
    pdataPing_rightImg = saved->pdataPing_rightImg;
    pdataPing_rightState = saved->pdataPing_rightState;
    pdataPing_rightMovelatency = saved->pdataPing_rightMovelatency;
    pdataState = saved->pdataState;
    pdataHurry_oddness = saved->pdataHurry_oddness;
    pdataSticking_step = saved->pdataSticking_step;
    pdataSleeping = saved->pdataSleeping;
    pdataPing_rightActionlatency = saved->pdataPing_rightActionlatency;
    pdataSticking_step_slowdown = saved->pdataSticking_step_slowdown;
    pinguinWin_roll_back_index = saved->pinguinWin_roll_back_index;
    pinguinLose_roll_back_index = saved->pinguinLose_roll_back_index;
    angle = saved->angle;
    time_1pgameextra = saved->time_1pgameextra;
    LevelsCurrent = saved->LevelsCurrent;
    next_bubble = saved->next_bubble;
    tobe_launched = saved->tobe_launched;
    sticked_bubbles.Reset();
    for(TInt idx=0; (idx<(TInt)(sizeof(saved->sticked_bubbles)/sizeof(*saved->sticked_bubbles))) && (saved->sticked_bubbles[idx].iColorIndex != BUBBLE_NULL); idx++){
      sticked_bubbles.Append(saved->sticked_bubbles[idx]);
    }
  }
  CleanupStack::PopAndDestroy( saved );
}

void CFrozenBubbleMain::save_game(void){
  // save ini file settings
  struct TSavedGameRecord *saved;
  saved = new(ELeave) TSavedGameRecord;
  CleanupStack::PushL( saved );
  saved->pdataScore = pdataScore;
  saved->pdataNewroot = pdataNewroot;
  saved->pdataNewroot_prelight = pdataNewroot_prelight;
  saved->pdataOddswap = pdataOddswap;
  saved->pdataHurry = pdataHurry;
  saved->pdataNewrootlevel = pdataNewrootlevel;
  saved->pdataNewroot_prelight_step = pdataNewroot_prelight_step;
  saved->pdataPing_rightImg = pdataPing_rightImg;
  saved->pdataPing_rightState = pdataPing_rightState;
  saved->pdataPing_rightMovelatency = pdataPing_rightMovelatency;
  saved->pdataState = pdataState;
  saved->pdataHurry_oddness = pdataHurry_oddness;
  saved->pdataSticking_step = pdataSticking_step;
  saved->pdataSleeping = pdataSleeping;
  saved->pdataPing_rightActionlatency = pdataPing_rightActionlatency;
  saved->pdataSticking_step_slowdown = pdataSticking_step_slowdown;
  saved->pinguinWin_roll_back_index = pinguinWin_roll_back_index;
  saved->pinguinLose_roll_back_index = pinguinLose_roll_back_index;
  saved->angle = angle;
  saved->time_1pgameextra = time_1pgameextra;
  saved->LevelsCurrent = LevelsCurrent;
  saved->next_bubble = next_bubble;
  saved->tobe_launched = tobe_launched;
  TInt idx=0;
  for(; idx<sticked_bubbles.Count(); idx++)
    saved->sticked_bubbles[idx] = sticked_bubbles[idx];
  for(; idx<(TInt)(sizeof(saved->sticked_bubbles)/sizeof(*saved->sticked_bubbles)); idx++)
    saved->sticked_bubbles[idx].iColorIndex = BUBBLE_NULL;
  Actions->WriteIniFile(uidSavedGame, saved, sizeof(*saved));
  CleanupStack::PopAndDestroy( saved );
}

void CFrozenBubbleMain::load_hiscores(void){
  THighScoreRecord buf[HIGHSCORELIMIT];
  TInt i;
  for(i=0; i<HIGHSCORELIMIT; i++)
    buf[i].level = 0;
  TUint size = sizeof(buf);
  if( Actions->ReadIniFile(uidHiscores, buf, size) ){
    highscores.Reset();
    for(i=0; i<HIGHSCORELIMIT && buf[i].level!=0; i++)
      highscores.Append( buf[i] );
  }
}

void CFrozenBubbleMain::save_hiscores(void){
  THighScoreRecord buf[HIGHSCORELIMIT];
  TInt i;
  for(i=0; i<HIGHSCORELIMIT; i++){
    if( i<highscores.Count() )
      buf[i] = highscores[i];
    else
      buf[i].level = 0;
  }
  Actions->WriteIniFile(uidHiscores, buf, sizeof(buf));
}

TInt CFrozenBubbleMain::HighScoreSort(const THighScoreRecord &a, const THighScoreRecord &b){
  if(a.level != b.level)
    return b.level - a.level;
  return a.seconds - b.seconds;
}

TInt CFrozenBubbleMain::BubbleCXOrder(const TBubble &a, const TBubble &b){
  return a.cx - b.cx;
}

void CFrozenBubbleMain::play_sound(TSample& sample){
  // Play the sample
  if( bSoundEnabled )
    smxMain->Play(sample, iActiveChannel, 8000, 64);
  
  if( ++iActiveChannel == 16 )
    iActiveChannel = 0;
}

void CFrozenBubbleMain::play_music(const TDesC8& aMusicName)
{
	if( bSoundEnabled )
	{
		TBuf8<256> filename;
		filename.Copy(iPrivatePath);
		filename.Append(aMusicName);
		smxMain->PlayMusic((const char*) filename.PtrZ());
	}
}

CFbsBitmap* CFrozenBubbleMain::ConvertBitmapL(CFbsBitmap *original, TDisplayMode mode){
  // Create a new bitmap, graphics device and context 
  CFbsBitmap* newBitmap = new ( ELeave ) CFbsBitmap();
  CleanupStack::PushL( newBitmap ); 
  newBitmap->Create( original->SizeInPixels(), mode ); 
  CFbsBitmapDevice* graphicsDevice = CFbsBitmapDevice::NewL( newBitmap ); 
  CleanupStack::PushL( graphicsDevice ); 
  CFbsBitGc* graphicsContext; 
  User::LeaveIfError( graphicsDevice->CreateContext( graphicsContext ) ); 
  // Blit the loaded bitmap to the new bitmap (the actual conversion) 
  graphicsContext->BitBlt( ptZero, original ); 
  CleanupStack::Pop(2);
  delete graphicsContext;
  delete graphicsDevice; 
  delete original; 
  return newBitmap;
}

void CFrozenBubbleMain::InitializeBitmap(CExtendedBitmap** iBmp, CFbsBitmap* iBitmap, TBool iDisposeBitmap, CFbsBitmap* iAlpha, TBool iDisposeAlpha){
  *iBmp = new(ELeave) CExtendedBitmap();
  CleanupStack::PushL( *iBmp );
  if( iBitmap ){
    (*iBmp)->SetBitmap( iBitmap );
    if(iDisposeBitmap)
      delete iBitmap;
  }
  if( iAlpha ){
    (*iBmp)->SetBitmapAlpha( iAlpha );
    if(iDisposeAlpha)
      delete iAlpha;
  }
  CleanupStack::Pop(*iBmp);
}

void CFrozenBubbleMain::AddBrightness(CPrintableExtendedBitmap &imgToBrighten){
  CFbsBitmap* iEmptyBitmap = new ( ELeave ) CFbsBitmap();
  CleanupStack::PushL( iEmptyBitmap );

  User::LeaveIfError( iEmptyBitmap->Create( imgToBrighten.GetBitmap()->SizeInPixels(), EikonEnv->DefaultDisplayMode() ) );
  CExtendedBitmap *imgBrightWhite = new (ELeave) CExtendedBitmap();
  CleanupStack::PushL( imgBrightWhite );

  CFbsBitmapDevice *iGD = CFbsBitmapDevice::NewL( iEmptyBitmap );
  CleanupStack::PushL( iGD );
  CFbsBitGc *iGC;
  User::LeaveIfError( iGD->CreateContext( iGC ) );

  iGC->SetPenStyle( CGraphicsContext::ENullPen );
  iGC->SetBrushColor( KRgbWhite );
  iGC->SetBrushStyle( CGraphicsContext::ESolidBrush );
  iGC->DrawRect( TRect(ptZero, iEmptyBitmap->SizeInPixels()) );
  iGC->SetBrushStyle( CGraphicsContext::ENullBrush );

  imgBrightWhite->SetBitmap( iEmptyBitmap );
  
  delete iGC;
  CleanupStack::PopAndDestroy( iGD );

  User::LeaveIfError( iEmptyBitmap->Create( imgToBrighten.GetBitmap()->SizeInPixels(), EGray256 ) );

  iGD = CFbsBitmapDevice::NewL( iEmptyBitmap );
  CleanupStack::PushL( iGD );
  User::LeaveIfError( iGD->CreateContext( iGC ) );

  iGC->SetPenStyle( CGraphicsContext::ENullPen );
  iGC->SetBrushColor( KRgbGray );
  iGC->SetBrushStyle( CGraphicsContext::ESolidBrush );
  iGC->DrawRect( TRect(ptZero, iEmptyBitmap->SizeInPixels()) );
  iGC->SetBrushStyle( CGraphicsContext::ENullBrush );

  imgBrightWhite->SetBitmapAlpha( iEmptyBitmap );

  delete iGC;
  CleanupStack::PopAndDestroy( iGD );

  imgBrightWhite->BitBltMasked(&imgToBrighten.getGraphicsContext(), ptZero);

  CleanupStack::PopAndDestroy( imgBrightWhite );
  CleanupStack::PopAndDestroy( iEmptyBitmap );
}

TUint CFrozenBubbleMain::getAppTicks(void){
  TTime now;
  now.HomeTime();
  TTimeIntervalMicroSeconds msecs = now.MicroSecondsFrom(total_time);
  
  return msecs.Int64()/1000;
}

void CFrozenBubbleMain::InitTimer(void){
  time_1pgame = getAppTicks();
  time_1pgameextra = 0;
}

void CFrozenBubbleMain::ResetTimeStamp(void){
  time_1pgame = getAppTicks();
}

void CFrozenBubbleMain::CalcTimer(void){
  time_1pgameextra += getAppTicks()-time_1pgame;
  time_1pgame = getAppTicks();
}

void CFrozenBubbleMain::setConstants(){
  // Constants from POS variable
  POSleft_limit  = 32;
  POSright_limit = 206;//143;
  POSpinguin.SetXY(114, 290);//74, 188);
  POSscoresx = 28;
  POSinit_top_limit = 30;//20;
  POSinitial_bubble_y = 258;//168;
  POSnext_bubble.SetXY(108, 293);//49, 190);
  POSmalus_y = 177;
  POShurry.SetXY(10, 120);
  POScenterpanel.SetXY(11, 110);
  POSpause_clip.SetXY(85, 141);
  POSscoresy = 282;//184;
  POScompressor_xpos = 120;//88;
  POStop_limit = POSinit_top_limit;
  BUBBLE_SIZE = 22;//14;
  ROW_SIZE    = BUBBLE_SIZE * 7/8;
  // other
  PI = 3.1415926535897932384626433832795028841972;
}

TInt64 CFrozenBubbleMain::getRandomSeed(void){
  TTime now;
  now.HomeTime();
  return now.Int64();
}

void CFrozenBubbleMain::setGameLoop(TInt (*iFunction)(TAny*)){
  sccLastKeyEvent = EStdKeyNull;
  sccPressedButton = EStdKeyNull;
  if( iFunction == maingame_ )
    global_state = GS_GAME;
  else if( iFunction == ask_nameloop_ )
    global_state = GS_NAME;
  else if( iFunction == highscoreloop_ )
    global_state = GS_HIGHSCORES;
  else if( iFunction == ask_keysloop_ )
    global_state = GS_KEYS;
  else if( iFunction == menu_loop_ )
    global_state = GS_MENU;
  else if( iFunction == dopause_loop_ )
    global_state = GS_PAUSE;
//  else if( iFunction == loadgame_loop_ )
//    global_state = GS_LOADGAME;
  prdGameLoop->Cancel();
  if(iFunction){
    bFirstLoopIteration = ETrue;
    prdGameLoop->Start(5000/*8000000*/, TARGET_ANIM_SPEED*1000/*35714*/, TCallBack(iFunction, this));
  }
}

TReal CFrozenBubbleMain::frand(TReal max){
  return Math::FRand(iRandomSeed) * max;
}

TInt CFrozenBubbleMain::rand(TInt dividor){
  return Math::Rand(iRandomSeed) % dividor;
}

// Bubble classes

TBubbleBase::TBubbleBase(void){
  iColorIndex = BUBBLE_NULL;
}

TBubble::TBubble(void):TBubbleBase(){
  chaindestx = -1;
  for(TUint idx=0; idx<sizeof(neighbours)/sizeof(*neighbours); idx++)
      neighbours[idx].iColorIndex = BUBBLE_NULL;    
}

TInt TBubble::addNeighbour(const TBubbleBase &bubble){
  for(TInt idx=0; idx<6; idx++)
    if(neighbours[idx].iColorIndex == BUBBLE_NULL){
      neighbours[idx] = bubble;
      return idx;
    } else if((neighbours[idx].iColorIndex == bubble.iColorIndex) &&
              (neighbours[idx].cx == bubble.cx) &&
              (neighbours[idx].cy == bubble.cy))
      return idx;
  return -1;
}

TInt TBubble::neighboursCount(void){
  TInt result=0;
  for(TInt idx=0; idx<6; idx++)
    if(neighbours[idx].iColorIndex != BUBBLE_NULL)
      result++;
  return result;
}

void TBubble::removeNeighbour(TInt idx){
  neighbours[idx].iColorIndex = BUBBLE_NULL;
  for(TInt i=idx+1; i<6; i++)
    if(neighbours[i].iColorIndex != BUBBLE_NULL){
      neighbours[i-1] = neighbours[i];
      neighbours[i].iColorIndex = BUBBLE_NULL;
    }
}

